# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

from typing import Iterable, Optional, Union

import click
from abutils import Sequence

from ..core.abstar import run as _run

# from ..core.germline import (
#     build_germdb_from_igdiscover as _build_germdb_from_igdiscover,
# )
from ..core.germline import build_germline_database as _build_germline_database
from ..utils.callbacks import HiddenClickOption, parse_dict_from_string


@click.group()
def cli():
    pass


# cli.add_command(build_germline_database)


@cli.command()
@click.argument(
    "input_path",
    type=str,
    # help="Path to a FASTA/Q file or a directory of FASTA/Q files. Gzip-compressed files are supported.",
)
@click.argument(
    "project_path",
    type=str,
    # help="Path to a directory in which tmp, log and output files will be deposited",
)
@click.option(
    "--germline_database",
    type=str,
    show_default=True,
    default="human",
    help="Name of the germline database to be used for assignment/annotation",
)
@click.option(
    "--receptor",
    type=click.Choice(["bcr", "tcr"], case_sensitive=False),
    show_default=True,
    default="bcr",
    help="Name of the receptor to be used for assignment/annotation",
)
@click.option(
    "-o",
    "--output_format",
    type=click.Choice(["airr", "parquet"], case_sensitive=False),
    multiple=True,
    show_default=True,
    default=["airr"],
    help="Format of the output files",
)
@click.option(
    "--umi_pattern",
    type=str,
    default=None,
    help="Pattern to match for extracting the UMI sequence, or name of a built-in pattern",
)
@click.option(
    "--umi_length",
    type=int,
    default=None,
    help="Length of the UMI sequence to extract. If positive, the UMI will be parsed from the start of the sequence. If negative, the UMI will be parsed from the end of the sequence.",
)
@click.option(
    "-m",
    "--merge",
    is_flag=True,
    default=False,
    help="Whether to merge FASTQ files prior to assignment",
)
@click.option(
    "--merge_kwargs",
    type=str,
    callback=parse_dict_from_string,
    default=None,
    help="Keyword arguments to pass to the merge_fastqs function. Format must be 'key1=val1,key2=val2'",
)
@click.option(
    "--interleaved_fastq",
    is_flag=True,
    default=False,
    help="Whether the input FASTQ files are interleaved",
)
@click.option(
    "-c",
    "--chunksize",
    type=int,
    show_default=True,
    default=500,
    help="Number of sequences to process at a time",
)
@click.option(
    "--mmseqs_chunksize",
    type=int,
    show_default=True,
    default=1e6,
    help="Number of sequences to process at a time for MMseqs2 searches (VDJC assignment)",
)
@click.option(
    "--mmseqs_threads",
    type=int,
    show_default=True,
    default=None,
    help="Number of threads to use for MMseqs2 searches (VDJC assignment)",
)
@click.option(
    "--n_processes",
    type=int,
    default=None,
    help="Number of processes to use for annotation",
)
@click.option(
    "--verbose/--quiet",
    default=True,
    help="Whether to print verbose output",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Whether to run in debug mode, which results in temporary files being retained and additional logging.",
)
@click.option(
    "--copy_inputs_to_project",
    cls=HiddenClickOption,
    is_flag=True,
    default=True,
)
@click.option(
    "--started_from_cli",
    cls=HiddenClickOption,
    is_flag=True,
    default=True,
)
def run(
    input_path: Union[str, Sequence, Iterable[Sequence]],
    project_path: Optional[str] = None,
    germline_database: str = "human",
    receptor: str = "bcr",
    output_format: Union[str, Iterable[str]] = "airr",
    umi_pattern: Optional[str] = None,
    umi_length: Optional[int] = None,
    merge: bool = False,
    merge_kwargs: Optional[dict] = None,
    interleaved_fastq: bool = False,
    chunksize: int = 500,
    mmseqs_chunksize: int = 1e6,
    mmseqs_threads: Optional[int] = None,
    n_processes: Optional[int] = None,
    copy_inputs_to_project: bool = False,
    verbose: bool = False,
    started_from_cli: bool = False,
    debug: bool = False,
) -> Optional[Union[Iterable[Sequence], Sequence]]:
    """
    Annotate antibody or TCR sequences.

    \b
    command line arguments:
      INPUT_PATH can be a FASTA/Q file or a directory of FASTA/Q files.
      PROJECT_PATH is the path to a directory in which tmp, log and output files will be deposited.
    """
    return _run(
        sequences=input_path,
        project_path=project_path,
        germline_database=germline_database,
        receptor=receptor,
        output_format=output_format,
        umi_pattern=umi_pattern,
        umi_length=umi_length,
        merge=merge,
        merge_kwargs=merge_kwargs,
        interleaved_fastq=interleaved_fastq,
        chunksize=chunksize,
        mmseqs_chunksize=mmseqs_chunksize,
        mmseqs_threads=mmseqs_threads,
        n_processes=n_processes,
        copy_inputs_to_project=copy_inputs_to_project,
        verbose=verbose,
        started_from_cli=started_from_cli,
        debug=debug,
    )


@cli.command(name="build_germline_database")
@click.argument(
    "name",
    type=str,
    required=True,
)
@click.option(
    "-f",
    "--fasta",
    "fasta_files",
    type=str,
    multiple=True,
    default=None,
    help="Path to a FASTA-formatted file containing gapped germline gene sequences. Can be used multiple times to provide multiple files.",
)
@click.option(
    "-j",
    "--json",
    "json_files",
    type=str,
    multiple=True,
    default=None,
    help="Path to a JSON-formatted file containing germline gene sequences. Can be used multiple times to provide multiple files.",
)
@click.option(
    "-c",
    "--constant",
    "constant_files",
    type=str,
    multiple=True,
    default=None,
    help="Path to a FASTA-formatted file containing gapped constant region sequences. Can be used multiple times to provide multiple files.",
)
@click.option(
    "-m",
    "--manifest",
    type=str,
    show_default=True,
    default=None,
    help="Path to a plain-text file containing information about the germline database. Format is not important, but this is the place for optional inforamtion like the origin of the germline database, the date of download, etc.",
)
@click.option(
    "-r",
    "--receptor",
    type=click.Choice(["bcr", "tcr"], case_sensitive=False),
    show_default=True,
    default="bcr",
    help="Receptor type",
)
@click.option(
    "-l",
    "--location",
    type=str,
    default=None,
    help="Location into which the new germline databases will be deposited. This option is provided primarily to test database creation without overwriting current databases of the same name.",
)
@click.option(
    "--reference",
    type=str,
    default="human",
    help="Reference species to use for adding IMGT gaps. Only used if ungapped VDJ sequences are provided.",
)
@click.option(
    "--include_species_in_name/--exclude_species_from_name",
    default=True,
    help="Whether to include the species in the name of each sequence, like so: 'IGHV1-2*02__homo_sapiens'. Only applies if input data is JSON-formatted or FASTA-formatted with species names already included in the sequence names.",
)
@click.option(
    "--verbose/--quiet",
    default=True,
    help="Print verbose output",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Run in debug mode, which will print more verbose information including stdout/stderr from command line tools.",
)
def build_germline_database(
    name: str,
    fasta_files: Optional[Union[str, Iterable[str]]] = None,
    json_files: Optional[Union[str, Iterable[str]]] = None,
    constant_files: Optional[Union[str, Iterable[str]]] = None,
    receptor: str = "bcr",
    manifest: Optional[str] = None,
    include_species_in_name: bool = True,
    location: Optional[str] = None,
    reference: str = "human",
    verbose: bool = True,
    debug: bool = False,
):
    """
    Build a custom germline database.
    """
    return _build_germline_database(
        name=name,
        fastas=fasta_files,
        jsons=json_files,
        constants=constant_files,
        receptor=receptor,
        manifest=manifest,
        include_species_in_name=include_species_in_name,
        location=location,
        reference=reference,
        verbose=verbose,
        debug=debug,
    )


# @cli.command()
# @click.argument(
#     "name",
#     type=str,
#     required=True,
# )
# @click.option(
#     "-i",
#     "--igdiscover_output",
#     type=str,
#     default=None,
#     help="Path to a directory containing the IgDiscover ouput. Usually myExperiment/final/database/",
# )
# @click.option(
#     "-c",
#     "--constants",
#     type=str,
#     default=None,
#     help="Path to a FASTA-formatted file containing gapped constant region sequences.",
# )
# @click.option(
#     "-r",
#     "--receptor",
#     type=click.Choice(["bcr", "tcr"], case_sensitive=False),
#     show_default=True,
#     default="bcr",
#     help="Receptor type",
# )
# @click.option(
#     "-s",
#     "--species",
#     type=str,
#     show_default=True,
#     default="human",
#     help="Name of the germline database to be used as source",
# )
# @click.option(
#     "-l",
#     "--location",
#     type=str,
#     default=None,
#     help="Location into which the new germline databases will be deposited. This option is provided primarily to test database creation without overwriting current databases of the same name.",
# )
# @click.option(
#     "--verbose/--quiet",
#     default=True,
#     help="Print verbose output",
# )
# @click.option(
#     "--debug",
#     is_flag=True,
#     default=False,
#     help="Run in debug mode, which will print more verbose information including stdout/stderr from command line tools.",
# )
# def build_germdb_from_igdiscover(
#     name: str,
#     igdiscover_output: str,
#     constants: Optional[str] = None,
#     receptor: str = "bcr",
#     species: str = "human",
#     location: Optional[str] = None,
#     verbose: bool = True,
#     debug: bool = False,
# ):
#     return _build_germdb_from_igdiscover(
#         name=name,
#         igdiscover_output=igdiscover_output,
#         constants=constants,
#         receptor=receptor,
#         species=species,
#         location=location,
#         verbose=verbose,
#         debug=debug,
#     )



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

from .annotation.germline import *
from .core.germline import *


#  TODO:
#
#  should think about the pros/cons of building a single, "combined" germline database per species that
#  contains both BCR and TCR germline segments
#
#  the most obvious con is the fact that OGRDB only has BCR data, so combined databases would need to be
#  mixture of IMGT and OGRDB datasets -- maybe this just needs some more time and OGRDB will eventually get TCR data?
#
#  functions:
#    - build_database() -- has `schema` kwarg, which can be either "ogrdb" or "imgt"
#    - build_database_ogrdb()
#       - should use their JSON files by default, but constant regions may need to be FASTA (and come from IMGT)
#    - build_database_imgt()
#
#    - show_manifest()
#       - prints the manifest for a given germline database



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


from .antibody import Antibody


def assess_productivity(ab: Antibody) -> Antibody:
    """
    Checks whether an Antibody is productive and annotates any
    productivity issues.

    Parameters
    ----------

    ab : Antibody
        Antibody object to update with annotation information. The following ``Antibody``
        properties are updated:

        - ``productive``
        - ``productivity_issues``

        The following ``Antibody`` properties must be populated:

        - ``sequence``
        - ``sequence_aa``
        - ``v_call``
        - ``j_call``
        - ``junction_aa``
    """
    # scan for issues
    if "*" in ab.sequence_aa:
        ab.productivity_issues.append("stop codon(s)")
        ab.stop_codon = True
    if "N" in ab.sequence:
        ab.productivity_issues.append("ambiguous nucleotide(s)")
    if ab.v_call[:3] != ab.j_call[:3]:
        ab.productivity_issues.append(
            f"V/J locus mismatch ({ab.v_call} and {ab.j_call})"
        )
    if ab.junction_aa:
        # conserved C at the start of the junction
        if ab.junction_aa[0] != "C":
            ab.productivity_issues.append("junction does not start with conserved C")
        # conserved W/F at the end of the junction
        if ab.locus is not None:
            if ab.locus.upper() in ["IGH", "TRA", "TRD"]:
                if ab.junction_aa[-1] != "W":
                    ab.productivity_issues.append(
                        "junction does not end with conserved W"
                    )
            else:
                if ab.junction_aa[-1] != "F":
                    ab.productivity_issues.append(
                        "junction does not end with conserved F"
                    )
        else:
            if ab.junction_aa[-1] not in ["F", "W"]:
                ab.productivity_issues.append(
                    "junction does not end with conserved W/F"
                )

    # flag sequences with issues as non-productive
    if ab.productivity_issues:
        ab.productive = False
    ab.productivity_issues = "|".join(ab.productivity_issues)

    return ab



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


from dataclasses import dataclass, field
from typing import Iterable, Optional, Union

# from ..utils.mixins import LoggingMixin
from abutils.tools.log import LoggingMixin


@dataclass
class Antibody(LoggingMixin):
    """
    Class for storing antibody annotation data.

    Includes the LoggingMixin, which provides methods for logging and exception handling.

    """

    # most useful info up front
    sequence_id: str = None
    v_gene: str = None
    d_gene: str = None
    j_gene: str = None
    c_gene: str = None
    cdr3_length: int = None
    junction_aa: str = None
    v_identity: float = None
    v_identity_aa: float = None
    d_identity: float = None
    d_identity_aa: float = None
    j_identity: float = None
    j_identity_aa: float = None
    productive: bool = True
    complete_vdj: bool = False

    # everything else
    sequence: str = None
    germline: str = None
    sequence_aa: str = None
    germline_aa: str = None
    sequence_vdjc: str = None
    germline_vdjc: str = None
    sequence_vdjc_aa: str = None
    germline_vdjc_aa: str = None
    sequence_lvdjc: str = None
    germline_lvdjc: str = None
    sequence_lvdjc_aa: str = None
    germline_lvdjc_aa: str = None
    sequence_gapped: str = None
    germline_gapped: str = None
    # sequence_gapped_aa: str = None  # indels in IMGT gaps can cause translation issues (if part of a codon is on one side of the IMGT gap and the rest is on the other side)
    # germline_gapped_aa: str = None
    sequence_vdjc_gapped: str = None
    germline_vdjc_gapped: str = None
    sequence_lvdjc_gapped: str = None
    germline_lvdjc_gapped: str = None
    sequence_alignment: str = None
    germline_alignment: str = None
    sequence_alignment_aa: str = None
    germline_alignment_aa: str = None
    umi: str = None
    quality: str = None
    locus: str = None
    species: str = None
    germline_database: str = None
    sequence_input: str = None
    sequence_oriented: str = None
    rev_comp: bool = False
    productivity_issues: list = field(default_factory=list)
    stop_codon: bool = False
    v_call: str = None
    v_score: float = None
    v_support: float = None
    v_cigar: str = None
    v_sequence: str = None
    v_germline: str = None
    v_sequence_gapped: str = None
    v_germline_gapped: str = None
    v_sequence_aa: str = None
    v_germline_aa: str = None
    v_sequence_gapped_aa: str = None
    v_germline_gapped_aa: str = None
    v_mutations: str = None
    v_mutations_aa: str = None
    v_mutation_count: int = None
    v_mutation_count_aa: int = None
    v_insertions: str = None
    v_deletions: str = None
    v_frameshift: bool = False
    frame: str = None
    d_call: str = None
    d_score: float = None
    d_support: float = None
    d_cigar: str = None
    d_sequence: str = None
    d_germline: str = None
    d_frame: str = None
    j_call: str = None
    j_score: float = None
    j_support: float = None
    j_cigar: str = None
    j_sequence: str = None
    j_germline: str = None
    c_call: str = None
    c_score: float = None
    c_support: float = None
    c_identity: float = None
    c_identity_aa: float = None
    c_cigar: str = None
    c_sequence: str = None
    c_germline: str = None
    c_sequence_gapped: str = None
    c_germline_gapped: str = None
    c_sequence_aa: str = None
    c_germline_aa: str = None
    c_sequence_gapped_aa: str = None
    c_germline_gapped_aa: str = None
    c_mutations: str = None
    c_mutations_aa: str = None
    c_mutation_count: int = None
    c_mutation_count_aa: int = None
    c_insertions: str = None
    c_deletions: str = None
    c_sequence_start: int = None
    c_sequence_end: int = None
    c_germline_start: int = None
    c_germline_end: int = None
    c_frame: int = None
    np1: str = None
    np2: str = None
    np1_length: int = None
    np2_length: int = None
    fwr1: str = None
    fwr1_aa: str = None
    cdr1: str = None
    cdr1_aa: str = None
    fwr2: str = None
    fwr2_aa: str = None
    cdr2: str = None
    cdr2_aa: str = None
    fwr3: str = None
    fwr3_aa: str = None
    cdr3: str = None
    cdr3_aa: str = None
    fwr4: str = None
    fwr4_aa: str = None
    junction: str = None
    v_sequence_start: int = None
    v_sequence_end: int = None
    v_germline_start: int = None
    v_germline_end: int = None
    j_sequence_start: int = None
    j_sequence_end: int = None
    j_germline_start: int = None
    j_germline_end: int = None

    def __post_init__(self):
        # establish the list of AIRR output fields
        self.airr_fields = list(self.__dict__.keys())

        # initialize the LoggingMixin
        super().__init__()

    def to_dict(
        self,
        include: Optional[Union[Iterable, str]] = None,
        exclude: Optional[Union[Iterable, str]] = None,
    ) -> dict:
        """
        Convert the Antibody object to a dictionary of annotations.

        Parameters:
        ----------
        include : Iterable or str, default: None
            Fields to include in the dictionary, in addition to the default fields.

        exclude : Iterable or str, default: None
            Fields to exclude from the dictionary.

        Returns:
        --------
        dict: The dictionary representation of the antibody.

        """
        airr_fields = self.airr_fields

        # excluded fields
        if exclude is not None:
            if isinstance(exclude, str):
                exclude = [exclude]
            airr_fields = [a for a in airr_fields if a not in exclude]

        # included fields
        if include is not None:
            if isinstance(include, str):
                include = [include]
            airr_fields.extend(include)

        return {k: self.__dict__.get(k, None) for k in airr_fields}



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT

import pytest

from ..annotation.indels import annotate_deletions, annotate_insertions

# =============================================
#                  FIXTURES
# =============================================


@pytest.fixture
def aligned_sequence_no_indels():
    return "ATGCATGC"


@pytest.fixture
def aligned_germline_no_indels():
    return "ATGCATGC"


@pytest.fixture
def gapped_germline_no_indels():
    return "ATGCATGC"


@pytest.fixture
def aligned_sequence_single_insertion():
    """Sequence with a single insertion (GGG) at position 4"""
    return "ATGCGGGATGC"


@pytest.fixture
def aligned_germline_single_insertion():
    """Germline with a gap corresponding to the insertion in aligned_sequence_single_insertion"""
    return "ATGC---ATGC"


@pytest.fixture
def gapped_germline_single_insertion():
    """IMGT-gapped germline with periods at positions 3, 4, and 5"""
    return "AT.GCA...TGC"


@pytest.fixture
def aligned_sequence_multi_insertion():
    """Sequence with multiple insertions: (GGG) at position 4 and (AA) at position 8"""
    return "ATGCGGGATAAATGC"


@pytest.fixture
def aligned_germline_multi_insertion():
    """Germline with gaps corresponding to the insertions in aligned_sequence_multi_insertion"""
    return "ATGC---AT--ATGC"


@pytest.fixture
def gapped_germline_multi_insertion():
    """IMGT-gapped germline with appropriate gaps"""
    return "AT.GCA...T..GATGC"


@pytest.fixture
def aligned_sequence_frameshift_insertion():
    """Sequence with a frameshift insertion (A) at position 4"""
    return "ATGCAATGC"


@pytest.fixture
def aligned_germline_frameshift_insertion():
    """Germline with a gap corresponding to the frameshift insertion"""
    return "ATGC-ATGC"


@pytest.fixture
def gapped_germline_frameshift_insertion():
    """IMGT-gapped germline with a period at position 5"""
    return "AT.GCA.TGC"


@pytest.fixture
def aligned_sequence_single_deletion():
    """Sequence with a single deletion at position 5 (deletion of T)"""
    return "ATGCAGC"


@pytest.fixture
def aligned_germline_single_deletion():
    """Germline with a character at the deletion position"""
    return "ATGCATGC"


@pytest.fixture
def gapped_germline_single_deletion():
    """IMGT-gapped germline"""
    return "AT.GCATGC"


@pytest.fixture
def aligned_sequence_multi_deletion():
    """Sequence with multiple deletions: positions 5-6 (deletion of TG)"""
    return "ATGCA--C"


@pytest.fixture
def aligned_germline_multi_deletion():
    """Germline with characters at the deletion positions"""
    return "ATGCATGC"


@pytest.fixture
def gapped_germline_multi_deletion():
    """IMGT-gapped germline"""
    return "AT.GCATGC"


@pytest.fixture
def aligned_sequence_frameshift_deletion():
    """Sequence with a frameshift deletion (single nucleotide)"""
    return "ATGCA-GC"


@pytest.fixture
def aligned_germline_frameshift_deletion():
    """Germline with character at the frameshift deletion position"""
    return "ATGCATGC"


@pytest.fixture
def gapped_germline_frameshift_deletion():
    """IMGT-gapped germline"""
    return "AT.GCATGC"


# =============================================
#             INSERTION TESTS
# =============================================


def test_annotate_insertions_no_insertions(
    aligned_sequence_no_indels,
    aligned_germline_no_indels,
    gapped_germline_no_indels,
):
    result = annotate_insertions(
        aligned_sequence=aligned_sequence_no_indels,
        aligned_germline=aligned_germline_no_indels,
        gapped_germline=gapped_germline_no_indels,
        germline_start=0,
    )
    assert result == ""


def test_annotate_insertions_single_insertion(
    aligned_sequence_single_insertion,
    aligned_germline_single_insertion,
    gapped_germline_single_insertion,
):
    result = annotate_insertions(
        aligned_sequence=aligned_sequence_single_insertion,
        aligned_germline=aligned_germline_single_insertion,
        gapped_germline=gapped_germline_single_insertion,
        germline_start=0,
    )
    assert result == "5:3>GGG"


def test_annotate_insertions_multi_insertion(
    aligned_sequence_multi_insertion,
    aligned_germline_multi_insertion,
    gapped_germline_multi_insertion,
):
    result = annotate_insertions(
        aligned_sequence=aligned_sequence_multi_insertion,
        aligned_germline=aligned_germline_multi_insertion,
        gapped_germline=gapped_germline_multi_insertion,
        germline_start=0,
    )
    assert "5:3>GGG" in result
    assert "10:2>AA" in result
    assert "|" in result  # Separator for multiple insertions
    assert result.count("|") == 1  # Only one separator


def test_annotate_insertions_frameshift(
    aligned_sequence_frameshift_insertion,
    aligned_germline_frameshift_insertion,
    gapped_germline_frameshift_insertion,
):
    result = annotate_insertions(
        aligned_sequence=aligned_sequence_frameshift_insertion,
        aligned_germline=aligned_germline_frameshift_insertion,
        gapped_germline=gapped_germline_frameshift_insertion,
        germline_start=0,
    )
    assert result == "5:1>A!"  # Note the ! marking frameshift


def test_annotate_insertions_germline_offset(
    aligned_sequence_single_insertion,
    aligned_germline_single_insertion,
    gapped_germline_single_insertion,
):
    """Test with non-zero germline_start parameter"""
    result = annotate_insertions(
        aligned_sequence=aligned_sequence_single_insertion,
        aligned_germline=aligned_germline_single_insertion,
        gapped_germline="AAAAAAAAAA" + gapped_germline_single_insertion,
        germline_start=10,  # Offset the germline start
    )
    assert result == "15:3>GGG"  # Position should be offset by 10


# =============================================
#              DELETION TESTS
# =============================================


def test_annotate_deletions_no_deletions(
    aligned_sequence_no_indels,
    aligned_germline_no_indels,
    gapped_germline_no_indels,
):
    result = annotate_deletions(
        aligned_sequence=aligned_sequence_no_indels,
        aligned_germline=aligned_germline_no_indels,
        gapped_germline=gapped_germline_no_indels,
        germline_start=0,
    )
    assert result == ""


def test_annotate_deletions_single_deletion(
    aligned_sequence_single_deletion,
    aligned_germline_single_deletion,
    gapped_germline_single_deletion,
):
    # Assuming the sequence has a deletion at position 5 (T)
    result = annotate_deletions(
        aligned_sequence="ATGCA-GC",  # Explicit dash to mark deletion
        aligned_germline=aligned_germline_single_deletion,
        gapped_germline=gapped_germline_single_deletion,
        germline_start=0,
    )
    assert result == "7:1>T!"


def test_annotate_deletions_multi_deletion(
    aligned_germline_multi_deletion,
    gapped_germline_multi_deletion,
):
    # Using explicit sequence with dashes to ensure proper alignment
    result = annotate_deletions(
        aligned_sequence="ATGCA--C",  # Two consecutive deletions
        aligned_germline=aligned_germline_multi_deletion,
        gapped_germline=gapped_germline_multi_deletion,
        germline_start=0,
    )
    assert result == "7-8:2>TG!"


def test_annotate_deletions_frameshift(
    aligned_germline_frameshift_deletion,
    gapped_germline_frameshift_deletion,
):
    # Using explicit sequence with dashes to ensure proper alignment
    result = annotate_deletions(
        aligned_sequence="ATGCA-GC",  # Single deletion causing frameshift
        aligned_germline=aligned_germline_frameshift_deletion,
        gapped_germline=gapped_germline_frameshift_deletion,
        germline_start=0,
    )
    assert result == "7:1>T!"  # Note the ! marking frameshift


def test_annotate_deletions_germline_offset(
    aligned_germline_single_deletion,
    gapped_germline_single_deletion,
):
    """Test with non-zero germline_start parameter"""
    result = annotate_deletions(
        aligned_sequence="ATGCA-GC",  # Explicit dash to mark deletion
        aligned_germline=aligned_germline_single_deletion,
        gapped_germline="AAAAAAAAAA" + gapped_germline_single_deletion,
        germline_start=10,  # Offset the germline start
    )
    assert result == "17:1>T!"  # Position should be offset by 10


def test_annotate_deletions_in_frame_deletion(
    aligned_germline_multi_deletion,
    gapped_germline_multi_deletion,
):
    # Using a sequence with a 3-nucleotide deletion (in-frame)
    result = annotate_deletions(
        aligned_sequence="ATG---GC",  # 3-nucleotide deletion
        aligned_germline="ATGCATGC",  # Original germline
        gapped_germline=gapped_germline_multi_deletion,
        germline_start=0,
    )
    assert result == "5-7:3>CAT"  # Should be in-frame (no ! mark)


def test_annotate_multiple_scattered_deletions():
    """Test multiple deletions that are not adjacent"""
    result = annotate_deletions(
        aligned_sequence="A-GC-TGC",  # Deletions at positions 1 and 4
        aligned_germline="ATGCATGC",
        gapped_germline="ATGCATGC",
        germline_start=0,
    )
    assert "2:1>T!" in result  # First deletion
    assert "5:1>A!" in result  # Second deletion
    assert "|" in result  # Separator
    assert result.count("|") == 1  # One separator for two deletions


# =============================================
#           COMBINED/EDGE CASE TESTS
# =============================================


def test_annotate_insertions_with_empty_strings():
    """Test with empty strings to ensure graceful handling"""
    result = annotate_insertions(
        aligned_sequence="",
        aligned_germline="",
        gapped_germline="",
        germline_start=0,
    )
    assert result == ""


def test_annotate_deletions_with_empty_strings():
    """Test with empty strings to ensure graceful handling"""
    result = annotate_deletions(
        aligned_sequence="",
        aligned_germline="",
        gapped_germline="",
        germline_start=0,
    )
    assert result == ""


def test_annotate_insertions_complex_case():
    """Test a more complex case with multiple insertions of varying lengths"""
    result = annotate_insertions(
        aligned_sequence="ATGCAGGGCATAACC",
        aligned_germline="ATGCA---CAT--CC",
        gapped_germline="ATGCA...C..AT..CC",
        germline_start=0,
    )
    assert "5:3>GGG" in result
    assert "13:2>AA" in result
    assert result.count("|") == 1  # One separator for two insertions


def test_annotate_deletions_complex_case():
    """Test a more complex case with multiple deletions of varying lengths"""
    result = annotate_deletions(
        aligned_sequence="AT--A-G--C",
        aligned_germline="ATGCATGAGC",
        gapped_germline="ATGCATGAGC",
        germline_start=0,
    )
    assert "3-4:2>GC" in result  # First deletion (2 nucleotides)
    assert "6:1>T" in result  # Second deletion (1 nucleotide)
    assert "8-9:2>AG" in result  # Third deletion (2 nucleotides)
    assert result.count("|") == 2  # Two separators for three deletions



#!/usr/bin/env python
# filename: germline.py

#
# Copyright (c) 2016 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import datetime
import difflib
import json
import os
import shutil
import subprocess as sp
import sys
from concurrent.futures import ThreadPoolExecutor
from typing import Iterable, Optional, Union

# from weakref import ref
import abutils
import parasail
from abutils import Sequence

# import click
# from natsort import natsorted
# from networkx import max_flow_min_cost
from tqdm.auto import tqdm

from ..gl import get_germline, get_germline_database_path
from ..utils import MATRIX_PATH

__all__ = ["build_germline_database"]

# ===============================
#
#   CUSTOM GERMLINE DATABASES
#
# ===============================


#  TODO: inputs/returns
#  --------------------
#
#  - inputs
#    - one or more FASTA or JSON files containing VDJ gene segments
#       - probably easiest to use a separate command-line flag for each file type that can be used multiple times -- "-f" for FASTA and "-j" for JSON
#       - no requirement for homogeneity of gene segments in the input files -- it doesn't have to be one file for V, one for J, etc
#       - combining heterogeneity with multiple JSON/FASTA files should make it much easier to build multi-species databases
#    - separate FASTA file(s) containing constant regions (since D genes and IgD genes both start with "IGHD")
#       - no need to support JSON at this point, because OGRDB doesn't have constant regions (yet)
#       - should be able to use a single flag that can be used multiple times -- "-c"?
#    - should also copy the unprocessed input data into a subdirectory ("raw"?) so that it's linked to the resulting germline database
#


#  TODO: overall workflow
#  ----------------------
#
#  - set up directory structure
#    - user provides a "project path" and we make tmp, log and output (airr and/or parquet) directories within it
#
#  - preprocessing steps
#    - for now, just sequence merging for paired FASTQ inputsm but maybe primer/adapter trimming for FASTA inputs in the future?
#    - all samples are processed before proceeding to assignment
#    - results get deposited in project/merged
#
#  - VDJC assignment
#    - processes an entire input/merged file in a single job (thanks, MMseqs!)
#    - output is a single parquet file, deposited in project/tmp/assignment
#    - log failed assignments in project/logs/assignment as a single txt file per sample
#
#  - split the assignment output into job files
#    - deposited in project/tmp/assignment
#    - once the splitting is done, should call cleanup() on the assigner to remove tmp files that are no longer needed
#
#  - annotation
#    - annotation jobs run in parallel (multiprocessing)
#    - outputs are deposited in project/tmp/annotation
#    - log failed annotations (and successful annotations, if debug=True) as separate failed/succeeeded files in project/log/annotation
#    - temporary output and log files are collected into separate lists so they can be merged and then deleted.
#
#  - create outputs
#    - concat output files (using polars) and write to single tsv (AIRR) and/or parquet files in project/airr and project/parquet
#    - concatenate log files into single failed/succeeded text files in project/logs/annotation
#    - remove all of the temporary output and log files


def build_germline_database(
    name: str,
    fastas: Optional[Union[str, Iterable[str]]] = None,
    jsons: Optional[Union[str, Iterable[str]]] = None,
    constants: Optional[Union[str, Iterable[str]]] = None,
    receptor: str = "bcr",
    manifest: Optional[str] = None,
    include_species_in_name: bool = True,
    location: Optional[str] = None,
    reference: str = "human",
    verbose: bool = True,
    debug: bool = False,
) -> None:
    """
    Builds a custom germline database.

    Parameters
    ----------
    name : str
        The name of the germline database. This will be how the database can be invoked when running ``abstar``.

        .. warning::
            Although custom databases are stored in a different location than built-in databases (and thus will not actually
            overwrite them on disk), custom databases are given priority over built-in databases when running ``abstar``.
            This means that a custom database named ``human`` will be used instead of the built-in ``human`` database.

    fastas : str | Iterable[str], optional
        The path to the FASTA file(s) containing the VDJ segments, either as a single file or a list of files.

    jsons : str | Iterable[str], optional
        The path to the JSON file(s) containing the VDJ segments, either as a single file or a list of files.

    constants : str | Iterable[str], optional
        The path to the FASTA file containing the constant regions, either as a single file or a list of files.

    receptor : str, default: "bcr"
        The type of receptor to build the database for.

    manifest : str, optional
        The path to the manifest file containing the metadata for the germline database.

    include_species_in_name : bool, default: True
        Whether to include the species name in the database name.

    location : str, optional
        The location of the germline database.

    reference : str, default: "human"
        The reference species to use for adding IMGT gaps. Only used if ungapped VDJ sequences are provided.

    verbose : bool, default: True
        Whether to print verbose output.

    debug : bool, default: False
        Whether to print debug output.

    Returns
    -------
    None
        The function does not return anything.

    """
    gapped_vdjs = []
    gapped_constants = []

    # set up database directory structure
    database_dir = get_database_directory(location, receptor)
    if check_for_existing_db(name, receptor, database_dir):
        confirm_overwrite_existing_db(name)
    database_dir = os.path.join(database_dir, name.lower())
    sub_dirs = make_db_directories(database_dir)
    raw_dir = sub_dirs["raw"]
    gapped_dir = sub_dirs["imgt_gapped"]
    ungapped_dir = sub_dirs["ungapped"]
    mmseqs_dir = sub_dirs["mmseqs"]

    # process FASTA-formatted VDJ segments
    if fastas:
        if verbose:
            print("processing FASTA-formatted VDJ segments")
        if isinstance(fastas, str):
            fastas = [fastas]
        for fasta in fastas:
            if not os.path.isfile(fasta):
                raise FileNotFoundError(f"The file {fasta} does not exist.")
            copy_to_raw(fasta, raw_dir)
            sequences = process_fasta(
                fasta_file=fasta, include_species_in_name=include_species_in_name
            )
            gapped_vdjs.extend(sequences)

    # process JSON-formatted VDJ segments
    if jsons:
        if verbose:
            print("processing JSON-formatted VDJ segments")
        if isinstance(jsons, str):
            jsons = [jsons]
        for _json in jsons:
            if not os.path.isfile(_json):
                raise FileNotFoundError(f"The file {_json} does not exist.")
            copy_to_raw(_json, raw_dir)
            sequences = process_json(
                _json, include_species_in_name=include_species_in_name
            )
            gapped_vdjs.extend(sequences)

    # add IMGT gapps (if they don't already exist)
    gapped_vdjs = add_imgt_gaps(gapped_vdjs, reference=reference)

    # process FASTA-formatted constant regions
    if constants:
        if verbose:
            print("processing FASTA-formatted constant regions")
        if isinstance(constants, str):
            constants = [constants]
        for constant in constants:
            if not os.path.isfile(constant):
                raise FileNotFoundError(f"The file {constant} does not exist.")
            copy_to_raw(constant, raw_dir)
            sequences = process_fasta(
                fasta_file=constant, include_species_in_name=False
            )
            gapped_constants.extend(sequences)

    # IMGT-gapped database
    if verbose:
        print("")
        print("building IMGT-gapped database")
    make_fasta_dbs(
        vdjs=gapped_vdjs,
        constants=gapped_constants,
        database_dir=gapped_dir,
        verbose=verbose,
    )

    # ungapped database
    if verbose:
        print("")
        print("building ungapped database")
    ungapped_vdjs = [
        Sequence(s.sequence.replace(".", ""), id=s.id) for s in gapped_vdjs
    ]
    ungapped_constants = [
        Sequence(s.sequence.replace(".", ""), id=s.id) for s in gapped_constants
    ]
    make_fasta_dbs(
        vdjs=ungapped_vdjs,
        constants=ungapped_constants,
        database_dir=ungapped_dir,
        verbose=verbose,
    )

    # MMseqs database
    if verbose:
        print("")
        print("building MMseqs database")
    make_mmseqs_dbs(
        database_dir=mmseqs_dir,
        ungapped_dir=ungapped_dir,
        verbose=verbose,
        debug=debug,
    )

    # manifest
    if manifest is not None:
        if verbose:
            print("")
            print("transferring manifest data")
        transfer_manifest_data(manifest, database_dir)


def add_imgt_gaps(
    germlines: abutils.Sequence | Iterable[abutils.Sequence], reference: str = "human"
) -> Iterable[abutils.Sequence]:
    """
    Add IMGT gaps to germline sequences.

    Parameters
    ----------
    germlines : abutils.Sequence | Iterable[abutils.Sequence]
        The germline sequences to add IMGT gaps to.

        .. note::
            Gaps are only added to V genes. D and J genes are always returned unchanged.

    reference : str, default: "human"
        The reference species to use for the IMGT gaps.

    Returns
    -------
    Iterable[abutils.Sequence]
        The germline sequences with IMGT gaps added. If any of the V gene sequences already
        have gaps, all sequences are returned unchanged.

    """
    if isinstance(germlines, abutils.Sequence):
        germlines = [germlines]
    # check for gaps -- if they're present, return the gapped sequences
    vgenes = [g for g in germlines if g.id[3] == "V"]
    others = [g for g in germlines if g.id[3] != "V"]
    if any(["." in v.sequence for v in vgenes]):
        return germlines

    # get IMGT-gapped germline V genes
    germs = get_germline("IGHV", germdb_name=reference, imgt_gapped=True)

    # build parasail matrix that more heavily penalizes mismatches to gaps
    matrix_file = os.path.join(MATRIX_PATH, "imgt_gapped.txt")
    matrix = parasail.Matrix(matrix_file)

    # add IMGT gaps
    gapped_vgenes = []
    for ungapped in vgenes:
        # find the best germline match
        alns = abutils.tl.semiglobal_alignment(
            ungapped, targets=germs, matrix=matrix, gap_open=-25
        )
        top_aln = alns[0]
        # build the gapped sequence
        gapped = ""
        for q, t in zip(top_aln.aligned_query, top_aln.aligned_target):
            if q == "-":
                gapped += "."
            else:
                gapped += q
        # remove trailing gaps, but keep any 5' gaps to preserve IMGT position information
        while True:
            if gapped[-1] == ".":
                gapped = gapped[:-1]
            else:
                break
        gapped_vgenes.append(abutils.Sequence(gapped, id=ungapped.id))
    return gapped_vgenes + others


# -------------------------
#   GENERATING CUSTOM DATABASE FROM IgDiscover
# -------------------------


# def build_germdb_from_igdiscover(
#     name: str,
#     igdiscover_output: str,
#     constants: Optional[str] = None,
#     receptor: str = "bcr",
#     species: str = "human",
#     location: Optional[str] = None,
#     verbose: bool = True,
#     debug: bool = False,
# ) -> None:
#     """
#     Builds a custom reference database using the output from IgDiscover

#     """

#     abutils.io.make_dir("/tmp/refs")

#     origin = get_germline_database_path(receptor=receptor, germdb_name=species)
#     shutil.copyfile(os.path.join(origin, "manifest.txt"), "/tmp/refs/manifest.txt")

#     with open("/tmp/refs/manifest.txt", "a") as f:
#         f.write("\n\n")
#         f.write("/!\\ CUSTOMIZED REFERENCE SPECIFIC TO DONOR /!\\\n\n")
#         f.write("Custom database modified with IgDiscover output\n")
#         f.write(f"Donor identifier: {name}\n")
#         f.write(f"Database created on {str(datetime.date.today())}\n")

#     files = abutils.io.list_files(igdiscover_output, extension="fasta")
#     for file_in in files:
#         filename = os.path.basename(file_in).lower()
#         file_out = os.path.join("/tmp/refs/", filename)

#         if "v" in filename:  # We only need to gap the V gene file
#             gapped_sequences = []
#             sequences = abutils.io.read_fasta(file_in)
#             v_refs = abutils.io.read_fasta(
#                 os.path.join(origin, "imgt_gapped", "v.fasta")
#             )
#             v_refs_hyphen = [
#                 Sequence(s.sequence.replace(".", "-"), id=s.id) for s in v_refs
#             ]

#             with ThreadPoolExecutor() as executor:
#                 if verbose:
#                     gapped_sequences = list(
#                         tqdm(
#                             executor.map(
#                                 lambda seq: pairwise_gap_sequence(
#                                     seq, v_refs_hyphen, verbose=verbose
#                                 ),
#                                 sequences,
#                             ),
#                             desc="Gapping new sequences...",
#                             total=len(sequences),
#                         )
#                     )

#                 else:
#                     gapped_sequences = list(
#                         executor.map(
#                             lambda seq: pairwise_gap_sequence(
#                                 seq, v_refs_hyphen, verbose=verbose
#                             ),
#                             sequences,
#                         ),
#                     )

#             with open(file_out, "w") as f:
#                 for s in gapped_sequences:
#                     f.write(s.fasta)
#                     f.write("\n")

#         else:  # D and J gene fils don't need to be gapped
#             shutil.copyfile(file_in, file_out)

#     # Constant genes are never part of the IgDiscover output, hence they need to be imported
#     if constants == None:
#         shutil.copyfile(
#             os.path.join(origin, "imgt_gapped", "c.fasta"), "/tmp/refs/c.fasta"
#         )

#     fastas = abutils.io.list_files("/tmp/refs/", extension="fasta")

#     # Building the germline_database using prepared files
#     build_germline_database(
#         name=name,
#         fastas=fastas,
#         constants=constants,
#         receptor=receptor,
#         manifest="/tmp/refs/manifest.txt",
#         include_species_in_name=False,
#         location=location,
#         verbose=verbose,
#         debug=debug,
#     )

#     if not debug:
#         shutil.rmtree("/tmp/refs")

#     return


# # def gap_sequence(sequence, reference, gaps='.'):
# #     to_align = [sequence, ] + reference
# #     aln = abutils.tools.alignment.mafft(sequences = to_align, mafft_bin='mafft')
# #     gapped_seq = [s for s in aln if s.id == sequence.id][0].sequence.replace("-", gaps)
# #     gapped = Sequence(gapped_seq, id=sequence.id)

# #     return gapped


# def pairwise_gap_sequence(sequence, references, gaps=".", verbose=False):
#     gene_name = sequence.id.split("_")[0]

#     # Getting the correct reference gene from IMGT to gap new sequence
#     try:
#         reference = [s for s in references if s.id == gene_name][0]
#     except:
#         match = difflib.get_close_matches(
#             gene_name, [r.id for r in references], n=1, cutoff=0.6
#         )[0]
#         if verbose:
#             print(
#                 f"No exact match for {gene_name}. Using {match} to perform alignement..."
#             )
#         reference = [s for s in references if s.id == match][0]

#     # Performing the alignment
#     aln = abutils.aln.semiglobal_alignment(
#         query=sequence, target=reference, mismatch=-5
#     )
#     gapped_seq = aln.aligned_query.replace("-", gaps)
#     gapped = Sequence(gapped_seq, id=sequence.id)

#     return gapped


# -------------------------
#   DATABASE DIRECTORIES
# -------------------------


def get_database_directory(receptor: str, db_location: Optional[str]) -> str:
    """
    Get the path to the receptor-level germline database directory.

    Parameters
    ----------
    db_location : Optional[str]
        The path to the addon directory. If not provided, the default location (~/.abstar/) will be used.

    receptor : str
        The receptor type.

    Returns
    -------
    str
        The path to the addon directory.

    """
    if db_location is not None:
        print("")
        print(
            "NOTE: You have selected a non-default location for the germline directory."
        )
        string = "abstar only looks in the default location (~/.abstar/) for user-created germline databases, "
        string += "so this database will not be used by abstar. The custom database location option is primarily "
        string += "provided so that users can test the database creation process without overwriting existing databases.\n"
        print(string)
        database_dir = db_location
    else:
        database_dir = os.path.expanduser("~/.abstar/germline_dbs")
    database_dir = os.path.join(database_dir, receptor.lower())
    abutils.io.make_dir(database_dir)
    return database_dir


def check_for_existing_db(
    name: str, receptor: str, location: Optional[str] = None
) -> bool:
    """
    Check if a germline database already exists in the addon directory.

    Parameters
    ----------
    name : str
        The name of the germline database.

    receptor : str
        The receptor type.

    location : Optional[str], default: None
        The path to a non-standard directory that may contain the germline database.

    Returns
    -------
    bool
        True if the germline database already exists, False otherwise.

    """
    if location is None:
        location = get_database_directory(receptor)
    dbs = [os.path.basename(d[0]) for d in os.walk(location)]
    if name.lower() in dbs:
        return True
    else:
        return False


def confirm_overwrite_existing_db(name: str) -> bool:
    """
    Confirm that the user wants to overwrite an existing germline database.

    Parameters
    ----------
    name : str
        The name of the germline database.

    Returns
    -------
    bool
        True if the user wants to continue, False otherwise.
    """
    print("")
    print(f"WARNING: A {name.lower()} germline database already exists.")
    print("Creating a new database with that name will overwrite the old one.")
    keep_going = input("Do you want to continue? [y/N]: ")
    if keep_going.lower() not in ["y", "yes"]:
        print("")
        print("Aborting germline database creation.")
        print("\n")
        sys.exit()
    else:
        print("")


def make_db_directories(database_dir: str) -> None:
    """
    Make the main directory for the germline database and the subdirectories for each file type.

    Parameters
    ----------
    database_dir : str
        The path to the database directory.

    """
    # make the main DB directory
    abutils.io.make_dir(database_dir)

    # make subdirectories
    subdirs = {}
    subdir_names = ["raw", "imgt_gapped", "ungapped", "mmseqs"]
    for subdir_name in subdir_names:
        subdir = os.path.join(database_dir, subdir_name)
        abutils.io.make_dir(subdir)
        subdirs[subdir_name] = subdir
    return subdirs


def transfer_manifest_data(manifest_file: str, database_dir: str) -> None:
    """
    Transfer manifest data to the new germline database.

    Parameters
    ----------
    manifest : str
        The path to the manifest file.

    addon_directory : str
        The path to the addon directory.

    dbname : str
        The name of the germline database.

    Returns
    -------
    str
        The path to the manifest file.

    """
    # read manifest data
    with open(manifest_file, "r") as f:
        manifest_data = f.read()

    # write to manifest file
    manifest_file = os.path.join(database_dir, "manifest.txt")
    with open(manifest_file, "w") as f:
        f.write(manifest_data)


def copy_to_raw(fasta: str, raw_dir: str) -> None:
    """
    Copy a FASTA file to the raw directory.

    Parameters
    ----------
    fasta : str
        The path to the input FASTA file.

    raw_dir : str
        The path to the raw directory.
    """
    shutil.copy(fasta, raw_dir)


# -------------------------
#   PROCESS INPUT FILES
# -------------------------


def process_fasta(
    fasta_file: str, include_species_in_name: bool = True
) -> Iterable[Sequence]:
    """
    Process a FASTA file and return a list of Sequence objects.
    """
    seqs = abutils.io.read_fasta(fasta_file)
    if not include_species_in_name:
        for s in seqs:
            s.id = s.id.split("__")[0]
    return seqs


def process_json(
    json_file: str, include_species_in_name: bool = True
) -> Iterable[Sequence]:
    """
    Process a JSON file and return a list of Sequence objects.
    """
    seqs = []
    with open(json_file, "r") as f:
        jdata = json.load(f)
    for entry in jdata["GermlineSet"][0]["allele_descriptions"]:
        name = entry["label"]
        if entry["sequence_type"] == "V":
            # V-genes are the only ones with IMGT-gapped sequences,
            # and it's in a different location than D/J sequences
            gapped = [
                d
                for d in entry["v_gene_delineations"]
                if d["delineation_scheme"] == "IMGT"
            ][0]["aligned_sequence"]
        else:
            gapped = entry["coding_sequence"]
        species = entry["species"]["label"].lower().replace(" ", "_")
        if include_species_in_name:
            name = f"{name}__{species}"
        seqs.append(Sequence(gapped, id=name))
    return seqs


# -------------------------
#     MAKE DATABASES
# -------------------------


def make_fasta_dbs(
    vdjs: Iterable[Sequence],
    constants: Iterable[Sequence],
    database_dir: str,
    verbose: bool = False,
) -> None:
    """
    Make the IMGT-gapped and ungapped FASTA databases.

    Parameters
    ----------
    vdjs : Iterable[Sequence]
        The VDJ genes.

    constants : Iterable[Sequence]
        The constant regions.

    database_dir : str
        The path to the database directory.

    verbose : bool, default: False
        Whether to print verbose output.
    """
    # VDJ genes
    for segment in ["V", "D", "J"]:
        if verbose:
            if segment == "V":
                print("  V", end="")
            else:
                print(f" | {segment}", end="")
        seqs = [s for s in vdjs if s.id[3] == segment]
        if seqs:
            abutils.io.to_fasta(
                seqs, os.path.join(database_dir, f"{segment.lower()}.fasta")
            )

    # constant regions
    if constants:
        if verbose:
            print(" | C")
        abutils.io.to_fasta(constants, os.path.join(database_dir, "c.fasta"))


def make_mmseqs_dbs(
    database_dir: str,
    ungapped_dir: str,
    verbose: bool = False,
    debug: bool = False,
) -> None:
    """
    Make the MMseqs2 database for a given segment.

    Parameters
    ----------
    input_file : str
        The path to the input (ungapped) FASTA file.

    addon_directory : str
        The path to the addon directory.

    segment : str
        The segment type.

    dbname : str
        The name of the germline database.

    verbose : bool, default: False
        Whether to print verbose output.

    debug : bool, default: False
        Whether to print debug output.

    """
    # VDJ genes
    for segment in ["V", "D", "J"]:
        if verbose:
            if segment == "V":
                print("  V", end="")
            else:
                print(f" | {segment}", end="")
        ungapped_file = os.path.join(ungapped_dir, f"{segment.lower()}.fasta")
        output_file = os.path.join(database_dir, f"{segment.lower()}")
        _make_mmseqs_db(ungapped_file, output_file, debug=debug)

    # constant regions
    segment = "C"
    ungapped_file = os.path.join(ungapped_dir, f"{segment.lower()}.fasta")
    if os.path.exists(ungapped_file):
        if verbose:
            print(f" | {segment}")
        output_file = os.path.join(database_dir, f"{segment.lower()}")
        _make_mmseqs_db(ungapped_file, output_file, debug=debug)
    elif verbose:
        print("")


def _make_mmseqs_db(input_file: str, output_file: str, debug: bool = False) -> None:
    """
    Make an MMseqs2 database.

    Parameters
    ----------
    input_file : str
        The path to the input file.

    output_file : str
        The path to the output file.

    debug : bool, default: False
        Whether to print debug output.
    """
    # create MMseqs2 database
    mmseqs_bin = abutils.bin.get_path("mmseqs")
    createdb_cmd = f"{mmseqs_bin} createdb {input_file} {output_file}"
    p = sp.Popen(createdb_cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    if debug:
        print(createdb_cmd)
        print(stdout)
        print(stderr)

    # create MMseqs2 index
    createindex_cmd = f"{mmseqs_bin} createindex {output_file} /tmp"
    p = sp.Popen(createindex_cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()
    if debug:
        print(createindex_cmd)
        print(stdout)
        print(stderr)


# def print_segment_info(segment: str, input_file: str) -> None:
#     """
#     Print information about the segment (variable, diversity, joining, constant).
#     """
#     seqs = abutils.io.read_fasta(input_file)
#     seg_string = "  " + segment.upper() + "  "
#     print("\n")
#     print("-" * len(seg_string))
#     print(seg_string)
#     print("-" * len(seg_string))
#     print(input_file)
#     print("input file contains {} sequences".format(len(seqs)))
#     print("")
#     print("Building germline databases:")


# def print_manifest_info(manifest: str) -> None:
#     seg_string = "  MANIFEST  "
#     print("\n")
#     print("-" * len(seg_string))
#     print(seg_string)
#     print("-" * len(seg_string))
#     print(manifest)
#     print("")
#     print("Transferring manifest data...")



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import re

from .antibody import Antibody
from .positions import get_gapped_position_from_raw

__all__ = [
    "annotate_insertions",
    "annotate_deletions",
]


def annotate_insertions(
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
) -> str:
    """
    Annotates non-templated insertions in the V gene region. The notation abstar uses for insertions is:

        ``"123:1>G!"`` (if the insertion causes a frameshift)
        ``"132:3>GGG"`` (if the insertion is in frame)

    In the case of multiple insertions, the insertions are separated by ``"|"`` characters, like so:

        ``"123:1>G!|132:3>GGG"``

    .. note::
        the position in abstar notation is the IMGT position after which the insertion occurs,
        so in the above example, the insertion ``"123:1>G!"`` would occur between positions 123 and 124

    Parameters
    ----------
    aligned_sequence : str
        The aligned query sequence for which insertions will be annotated.

    aligned_germline : str
        The aligned germline sequence.

    gapped_germline : str
        The IMGT-gapped germline sequence.

    germline_start : int
        The start position of the germline sequence in the alignment.

    Returns
    -------
    str
        A ``"|"``-separated string of non-templated insertions in the V gene region.
    """
    # annotate insertions
    insertions = []
    for i in re.finditer("-+", aligned_germline):
        raw_start = i.start()
        raw_end = i.end()
        length = raw_end - raw_start
        seq = aligned_sequence[raw_start:raw_end]

        # get the aligned germline position of the insertion start
        start = len(aligned_germline[:raw_start].replace("-", "")) + germline_start
        imgt_start = get_gapped_position_from_raw(start, gapped_germline)

        # format the insertion
        oof = "!" if length % 3 else ""
        insertions.append(f"{imgt_start}:{length}>{seq}{oof}")
    return "|".join(insertions)


def annotate_deletions(
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
) -> str:
    """
    Annotates non-templated deletions in the V gene region. The notation abstar uses for deletions is:

        ``123:1>G!`` (if only single nucleotide in length)
        ``123-124:2>GG!`` (if multi-nucleotide and causing a frameshift)
        ``123-125:3>GGG`` (if multi-nucleotide and in frame)

    In the case of multiple deletions, the deletions are separated by ``"|"`` characters, like so:

        ``"123:1>G!|132-134:3>GGG"``

    .. note::
        The position range in abstar notation denotes the IMGT position of the start and end of the deletion,
        inclusive. This conforms to IMGT's convention for deletion position numbering.

    Parameters
    ----------
    aligned_sequence : str
        The aligned query sequence for which insertions will be annotated.

    aligned_germline : str
        The aligned germline sequence.

    gapped_germline : str
        The IMGT-gapped germline sequence.

    germline_start : int
        The start position of the germline sequence in the alignment.

    Returns
    -------
    str
        A ``"|"``-separated string of non-templated deletions in the V gene region.
    """
    # annotate deletions
    deletions = []
    for i in re.finditer("-+", aligned_sequence):
        raw_start = i.start()
        raw_end = i.end()
        length = raw_end - raw_start
        seq = aligned_germline[raw_start:raw_end]

        # get the aligned germline position of the deletion start
        #
        # unlike insertions, we want the start position of the actual deletion,
        # which is why we have the +1 on the start (for insertions, we want the
        # position immediately preceeding the insertion)
        start = len(aligned_germline[:raw_start].replace("-", "")) + germline_start + 1
        imgt_start = get_gapped_position_from_raw(start, gapped_germline)
        imgt_end = imgt_start + length

        # format the deletion
        oof = "!" if length % 3 else ""
        if length > 1:
            pos_range = f"{imgt_start}-{imgt_end - 1}"  # end is inclusive
        else:
            pos_range = f"{imgt_start}"
        deletions.append(f"{pos_range}:{length}>{seq}{oof}")
    return "|".join(deletions)


# # ---------------------------------------
# #       VARIABLE REGION INDELS
# # ---------------------------------------


# def annotate_v_insertions(
#     aligned_sequence: str,
#     aligned_germline: str,
#     gapped_germline: str,
#     germline_start: int,
#     ab: Antibody,
# ) -> Antibody:
#     """
#     Annotates non-templated insertions in the V gene region. The notation abstar uses for insertions is:

#         ``"123:1>G!"`` (if the insertion causes a frameshift)
#         ``"132:3>GGG"`` (if the insertion is in frame)

#     In the case of multiple insertions, the insertions are separated by ``"|"`` characters, like so:

#         ``"123:1>G!|132:3>GGG"``

#     .. note::
#         the position in abstar notation is the IMGT position after which the insertion occurs,
#         so in the above example, the insertion ``"123:1>G!"`` would occur between positions 123 and 124

#     Parameters
#     ----------
#     aligned_sequence : str
#         The aligned query sequence for which insertions will be annotated.

#     aligned_germline : str
#         The aligned germline sequence.

#     gapped_germline : str
#         The IMGT-gapped germline sequence.

#     germline_start : int
#         The start position of the germline sequence in the alignment.

#     ab : Antibody
#         The ``Antibody`` object.

#         The following ``Antibody`` properties are updated:

#             - ``v_insertions``: a ``"|"``-separated string of non-templated insertions in the V gene region

#         This function does not require any ``Antibody`` properties to be set.

#     Returns
#     -------
#     Antibody
#         The updated ``Antibody`` object.

#     """
#     # annotate insertions
#     insertions = annotate_insertions(
#         aligned_sequence=aligned_sequence,
#         aligned_germline=aligned_germline,
#         gapped_germline=gapped_germline,
#         germline_start=germline_start,
#     )
#     ab.v_insertions = "|".join(insertions)
#     return ab


# def annotate_v_deletions(
#     aligned_sequence: str,
#     aligned_germline: str,
#     gapped_germline: str,
#     germline_start: int,
#     ab: Antibody,
# ) -> Antibody:
#     """
#     Annotates non-templated deletions in the V gene region. The notation abstar uses for deletions is:

#         ``123:1>G!`` (if only single nucleotide in length)
#         ``123-124:2>GG!`` (if multi-nucleotide and causing a frameshift)
#         ``123-125:3>GGG`` (if multi-nucleotide and in frame)

#     In the case of multiple deletions, the deletions are separated by ``"|"`` characters, like so:

#         ``"123:1>G!|132-134:3>GGG"``

#     .. note::
#         The position range in abstar notation denotes the IMGT position of the start and end of the deletion,
#         inclusive. This conforms to IMGT's convention for deletion position numbering.

#     Parameters
#     ----------
#     aligned_sequence : str
#         The aligned query sequence for which insertions will be annotated.

#     aligned_germline : str
#         The aligned germline sequence.

#     gapped_germline : str
#         The IMGT-gapped germline sequence.

#     germline_start : int
#         The start position of the germline sequence in the alignment.

#     ab : Antibody
#         The ``Antibody`` object.

#         The following ``Antibody`` properties are updated:

#             - ``v_insertions``: a ``"|"``-separated string of non-templated insertions in the V gene region

#         This function does not require any ``Antibody`` properties to be set.

#     Returns
#     -------
#     Antibody
#         The updated ``Antibody`` object.

#     """
#     # annotate deletions
#     deletions = annotate_deletions(
#         aligned_sequence=aligned_sequence,
#         aligned_germline=aligned_germline,
#         gapped_germline=gapped_germline,
#         germline_start=germline_start,
#     )
#     ab.v_deletions = "|".join(deletions)
#     return ab


# # ---------------------------------------
# #       CONSTANT REGION INDELS
# # ---------------------------------------


# def annotate_c_insertions(
#     aligned_sequence: str,
#     aligned_germline: str,
#     gapped_germline: str,
#     germline_start: int,
#     ab: Antibody,
# ) -> Antibody:
#     """
#     Annotates non-templated insertions in the V gene region. The notation abstar uses for insertions is:

#         ``"123:1>G!"`` (if the insertion causes a frameshift)
#         ``"132:3>GGG"`` (if the insertion is in frame)

#     In the case of multiple insertions, the insertions are separated by ``"|"`` characters, like so:

#         ``"123:1>G!|132:3>GGG"``

#     .. note::
#         the position in abstar notation is the IMGT position after which the insertion occurs,
#         so in the above example, the insertion ``"123:1>G!"`` would occur between positions 123 and 124

#     Parameters
#     ----------
#     aligned_sequence : str
#         The aligned query sequence for which insertions will be annotated.

#     aligned_germline : str
#         The aligned germline sequence.

#     gapped_germline : str
#         The IMGT-gapped germline sequence.

#     germline_start : int
#         The start position of the germline sequence in the alignment.

#     ab : Antibody
#         The ``Antibody`` object.

#         The following ``Antibody`` properties are updated:

#             - ``v_insertions``: a ``"|"``-separated string of non-templated insertions in the V gene region

#         This function does not require any ``Antibody`` properties to be set.

#     Returns
#     -------
#     Antibody
#         The updated ``Antibody`` object.

#     """
#     # annotate insertions
#     insertions = annotate_insertions(
#         aligned_sequence=aligned_sequence,
#         aligned_germline=aligned_germline,
#         gapped_germline=gapped_germline,
#         germline_start=germline_start,
#     )
#     ab.c_insertions = "|".join(insertions)
#     return ab


# def annotate_c_deletions(
#     aligned_sequence: str,
#     aligned_germline: str,
#     gapped_germline: str,
#     germline_start: int,
#     ab: Antibody,
# ) -> Antibody:
#     """
#     Annotates non-templated deletions in the V gene region. The notation abstar uses for deletions is:

#         ``123:1>G!`` (if only single nucleotide in length)
#         ``123-124:2>GG!`` (if multi-nucleotide and causing a frameshift)
#         ``123-125:3>GGG`` (if multi-nucleotide and in frame)

#     In the case of multiple deletions, the deletions are separated by ``"|"`` characters, like so:

#         ``"123:1>G!|132-134:3>GGG"``

#     .. note::
#         The position range in abstar notation denotes the IMGT position of the start and end of the deletion,
#         inclusive. This conforms to IMGT's convention for deletion position numbering.

#     Parameters
#     ----------
#     aligned_sequence : str
#         The aligned query sequence for which insertions will be annotated.

#     aligned_germline : str
#         The aligned germline sequence.

#     gapped_germline : str
#         The IMGT-gapped germline sequence.

#     germline_start : int
#         The start position of the germline sequence in the alignment.

#     ab : Antibody
#         The ``Antibody`` object.

#         The following ``Antibody`` properties are updated:

#             - ``v_insertions``: a ``"|"``-separated string of non-templated insertions in the V gene region

#         This function does not require any ``Antibody`` properties to be set.

#     Returns
#     -------
#     Antibody
#         The updated ``Antibody`` object.

#     """
#     # annotate deletions
#     deletions = annotate_deletions(
#         aligned_sequence=aligned_sequence,
#         aligned_germline=aligned_germline,
#         gapped_germline=gapped_germline,
#         germline_start=germline_start,
#     )
#     ab.c_deletions = "|".join(deletions)
#     return ab



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


__all__ = [
    "get_gapped_position_from_raw",
    "get_raw_position_from_gapped",
    "get_raw_position_from_aligned",
    "get_gapped_sequence",
]


def get_gapped_position_from_raw(
    position: int,
    gapped_germline: str,
) -> int:
    """
    Get the IMGT-gapped germline position from a raw (ungapped) germline position.

    .. note::
        This function uses 1-based indexing, since that's what IMGT uses for position numbering.

    Parameters
    ----------
    position : int
        The raw (ungapped) position to convert to a gapped position.

    gapped_germline : str
        The gapped germline sequence to convert the raw position to a gapped position.

    Returns
    -------
    int
        The IMGT-gapped position.

    """
    raw = 0
    gapped = 0
    for res in gapped_germline:
        if res == ".":
            gapped += 1
        else:
            raw += 1
            gapped += 1
        if raw == position:
            break
    return gapped


def get_raw_position_from_gapped(
    position: int,
    gapped_germline: str,
    sequence_start: int = 0,
    germline_start: int = 0,
) -> int:
    """
    Get the raw (ungapped) germline position from an IMGT-gapped germline position.

    .. note::
        This function uses 1-based indexing, since that's what IMGT uses for position numbering.

    Parameters
    ----------
    position : int
        The IMGT-gapped position to convert to a raw (ungapped) position.

    gapped_germline : str
        The gapped germline sequence to convert the gapped position to a raw position.

    sequence_start : int, default 0
        The start position of the query sequence in the alignment with germline.

    germline_start : int, default 0
        The start position of the germline sequence in the alignment with the query sequence.

    Returns
    -------
    int
        The raw (ungapped) position.

    """
    raw = len(gapped_germline[:position].replace(".", ""))
    raw -= germline_start
    raw -= sequence_start
    return max([0, raw])


def get_raw_position_from_aligned(
    position: int,
    aligned_sequence: str,
    aligned_reference: str,
) -> int:
    """
    Gets the aligned position in a sequence, given an aligned reference.

    Parameters
    ----------
    position : int
        The aligned position to convert to a raw (ungapped) position.

    aligned_sequence : str
        The aligned query sequence from which to convert an aligned position to a raw position.

    aligned_reference : str
        The aligned reference sequence.

    Returns
    -------
    int
        The raw (ungapped) position.
    """
    aligned_position = 0
    raw_position = 0
    for s, r in zip(aligned_sequence, aligned_reference):
        if r == "-":
            aligned_position += 1
        else:
            aligned_position += 1
            raw_position += 1
        if raw_position == position:
            return aligned_position


def get_gapped_sequence(
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
) -> str:
    """
    Get the gapped sequence from an aligned sequence and aligned germline.

    Parameters
    ----------
    aligned_sequence : str
        The aligned sequence to convert to a gapped sequence.

    aligned_germline : str
        The aligned germline sequence.

    gapped_germline : str
        The gapped germline sequence, which will be used to determine where gaps should be inserted.

    germline_start : int
        The start position of the germline sequence in the alignment with the query sequence.

    Returns
    -------
    str
        The gapped sequence.

    """
    gapped_seq = ""
    seq_pos = 0
    germ_pos = 0
    gap_pos = germline_start

    while gap_pos < len(gapped_germline):
        seq = aligned_sequence[seq_pos]
        germ = aligned_germline[germ_pos]
        gap = gapped_germline[gap_pos]
        if gap == "." and germ != "-":
            gapped_seq += gap
            gap_pos += 1
        else:
            gapped_seq += seq
            seq_pos += 1
            germ_pos += 1
            gap_pos += 1
        if seq_pos >= len(aligned_sequence):
            break

    return gapped_seq



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import polars as pl


class NoneDict(dict):
    """
    A dictionary that returns None for missing keys.
    """

    def __getitem__(self, key):
        return dict.get(self, key)


schema_dict = {
    "sequence_id": pl.String,
    "v_gene": pl.String,
    "d_gene": pl.String,
    "j_gene": pl.String,
    "c_gene": pl.String,
    "cdr3_length": pl.Int64,
    "junction_aa": pl.String,
    "v_identity": pl.Float64,
    "v_identity_aa": pl.Float64,
    "d_identity": pl.Float64,
    "d_identity_aa": pl.Float64,
    "j_identity": pl.Float64,
    "j_identity_aa": pl.Float64,
    "productive": pl.Boolean,
    "complete_vdj": pl.Boolean,
    "sequence": pl.String,
    "germline": pl.String,
    "sequence_vdjc": pl.String,
    "germline_vdjc": pl.String,
    "sequence_lvdjc": pl.String,
    "germline_lvdjc": pl.String,
    "sequence_aa": pl.String,
    "germline_aa": pl.String,
    "sequence_vdjc_aa": pl.String,
    "germline_vdjc_aa": pl.String,
    "sequence_lvdjc_aa": pl.String,
    "germline_lvdjc_aa": pl.String,
    "sequence_gapped": pl.String,
    "germline_gapped": pl.String,
    # "sequence_gapped_aa": pl.String,  # indels in IMGT gaps can cause translation issues (if part of a codon is on one side of the IMGT gap and the rest is on the other side)
    # "germline_gapped_aa": pl.String,
    "sequence_vdjc_gapped": pl.String,
    "germline_vdjc_gapped": pl.String,
    "sequence_lvdjc_gapped": pl.String,
    "germline_lvdjc_gapped": pl.String,
    "sequence_alignment": pl.String,
    "germline_alignment": pl.String,
    "sequence_alignment_aa": pl.String,
    "germline_alignment_aa": pl.String,
    "umi": pl.String,
    "quality": pl.String,
    "locus": pl.String,
    "species": pl.String,
    "germline_database": pl.String,
    "sequence_input": pl.String,
    "sequence_oriented": pl.String,
    "rev_comp": pl.Boolean,
    "productivity_issues": pl.String,
    "stop_codon": pl.Boolean,
    "v_call": pl.String,
    "v_score": pl.Float64,
    "v_support": pl.Float64,
    "v_cigar": pl.String,
    "v_sequence": pl.String,
    "v_germline": pl.String,
    "v_sequence_gapped": pl.String,
    "v_germline_gapped": pl.String,
    "v_sequence_aa": pl.String,
    "v_germline_aa": pl.String,
    # "v_sequence_gapped_aa": pl.String,
    # "v_germline_gapped_aa": pl.String,
    "v_mutations": pl.String,
    "v_mutations_aa": pl.String,
    "v_mutation_count": pl.Int64,
    "v_mutation_count_aa": pl.Int64,
    "v_insertions": pl.String,
    "v_deletions": pl.String,
    "v_frameshift": pl.Boolean,
    "v_frame": pl.Int64,
    "d_call": pl.String,
    "d_score": pl.Float64,
    "d_support": pl.Float64,
    "d_cigar": pl.String,
    "d_sequence": pl.String,
    "d_germline": pl.String,
    "d_frame": pl.Int64,
    "j_call": pl.String,
    "j_score": pl.Float64,
    "j_support": pl.Float64,
    "j_cigar": pl.String,
    "j_sequence": pl.String,
    "j_germline": pl.String,
    "c_call": pl.String,
    "c_score": pl.Float64,
    "c_support": pl.Float64,
    "c_identity": pl.Float64,
    "c_identity_aa": pl.Float64,
    "c_cigar": pl.String,
    "c_sequence": pl.String,
    "c_germline": pl.String,
    "c_sequence_gapped": pl.String,
    "c_germline_gapped": pl.String,
    "c_sequence_aa": pl.String,
    "c_germline_aa": pl.String,
    "c_sequence_gapped_aa": pl.String,
    "c_germline_gapped_aa": pl.String,
    "c_mutations": pl.String,
    "c_mutations_aa": pl.String,
    "c_mutation_count": pl.Int64,
    "c_mutation_count_aa": pl.Int64,
    "c_insertions": pl.String,
    "c_deletions": pl.String,
    "c_sequence_start": pl.Int64,
    "c_sequence_end": pl.Int64,
    "c_germline_start": pl.Int64,
    "c_germline_end": pl.Int64,
    "c_frame": pl.Int64,
    "np1": pl.String,
    "np2": pl.String,
    "np1_length": pl.Int64,
    "np2_length": pl.Int64,
    "fwr1": pl.String,
    "fwr1_aa": pl.String,
    "cdr1": pl.String,
    "cdr1_aa": pl.String,
    "fwr2": pl.String,
    "fwr2_aa": pl.String,
    "cdr2": pl.String,
    "cdr2_aa": pl.String,
    "fwr3": pl.String,
    "fwr3_aa": pl.String,
    "cdr3": pl.String,
    "cdr3_aa": pl.String,
    "fwr4": pl.String,
    "fwr4_aa": pl.String,
    "junction": pl.String,
    "v_sequence_start": pl.Int64,
    "v_sequence_end": pl.Int64,
    "v_germline_start": pl.Int64,
    "v_germline_end": pl.Int64,
    "j_sequence_start": pl.Int64,
    "j_sequence_end": pl.Int64,
    "j_germline_start": pl.Int64,
    "j_germline_end": pl.Int64,
}


OUTPUT_SCHEMA = NoneDict(schema_dict)



# Copyright (c) 2024 brineylab @ scripps
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import os

from setuptools import find_packages, setup

# the actual __version__ is read from version.py, not assigned directly
# this causes the linter to freak out when we mention __version__ in setup()
# to fix that, we fake assign it here
__version__ = None

# read version
version_file = os.path.join(os.path.dirname(__file__), "abstar", "version.py")
with open(version_file) as version:
    exec(version.read())

# read requirements
with open("requirements.txt") as reqs:
    requirements = reqs.read().splitlines()

# read long description
with open("README.md", encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name="abstar",
    version=__version__,
    author="Bryan Briney",
    author_email="briney@scripps.edu",
    description="Germline assignment and annotation of adaptive immune receptor repertoire (AIRR) data. Scalable from a single sequence to billions of sequences.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/brineylab/abstar",
    packages=find_packages(),
    include_package_data=True,
    # scripts=[
    #     "bin/abstar",
    #     # "bin/batch_mongoimport",
    #     # "bin/basespace_downloader",
    #     # "bin/build_abstar_germline_db",
    #     # "bin/make_basespace_credfile",
    # ],
    entry_points={
        "console_scripts": [
            "abstar=abstar.scripts.abstar:cli",
        ]
    },
    install_requires=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    python_requires=">=3.10",
)


# import os
# import sys

# try:
#     from setuptools import setup
# except ImportError:
#     sys.exit('ERROR: setuptools is required.\n')

# try: # for pip >= 10
#     from pip._internal.req import parse_requirements
# except ImportError: # for pip <= 9.0.3
#     from pip.req import parse_requirements
# # try:
# #     from pip.req import parse_requirements
# # except ImportError:
# #     sys.exit('ERROR: pip is required.\n')


# if os.environ.get('READTHEDOCS', None):
#     # Set empty install_requires to get install to work on readthedocs
#     install_requires = []
# else:
#     req_file = 'requirements.txt'
#     try:
#         reqs = parse_requirements(req_file, session=False)
#     except TypeError:
#         reqs = parse_requirements(req_file)
#     try:
#         install_requires = [str(r.req) for r in reqs]
#     except AttributeError:
#         install_requires = [str(r.requirement) for r in reqs]

# # read version
# exec(open('abstar/version.py').read())

# # read long description
# with open("README.md", "r") as fh:
#     long_description = fh.read()


# config = {
#     'name': 'abstar',
#     'version': __version__,
#     'author': 'Bryan Briney',
#     'author_email': 'briney@scripps.edu',
#     'description': 'VDJ assignment and antibody sequence annotation. Scalable from a single sequence to billions of sequences.',
#     'long_description': long_description,
#     'long_description_content_type': 'text/markdown',
#     'url': 'https://www.github.com/briney/abstar',
#     'install_requires': install_requires,
#     'packages': ['abstar'],
#     'scripts': ['bin/abstar',
#                 'bin/batch_mongoimport',
#                 'bin/basespace_downloader',
#                 'bin/build_abstar_germline_db',
#                 'bin/make_basespace_credfile',],
#     'include_package_data': True,
#     'classifiers': ['License :: OSI Approved :: MIT License',
#                     'Programming Language :: Python :: 3.8',
#                     'Programming Language :: Python :: 3.9',
#                     'Topic :: Scientific/Engineering :: Bio-Informatics']
# }

# setup(**config)



# #!usr/env/python
# # filename: test_integration.py

# import os
# import sys
# import tempfile

# import pytest
# from abutils.core.sequence import Sequence
# from Bio import SeqIO

# from ..core import abstar


# # def test_api_integration_bcr_1k():
# #     test_data = os.path.abspath('abstar/test_data/test_1k.fasta')
# #     with open(test_data, 'r') as f:
# #         test_seqs = [Sequence(s) for s in SeqIO.parse(f, 'fasta')]
# #     seqs = abstar.run(*test_seqs)
# #     assert len(seqs) == 999


# # def test_api_integration_bcr_hiv_bnab_hcs():
# #     test_data = os.path.abspath('abstar/test_data/test_hiv_bnab_hcs.fasta')
# #     with open(test_data, 'r') as f:
# #         test_seqs = [Sequence(s) for s in SeqIO.parse(f, 'fasta')]
# #     seqs = abstar.run(*test_seqs)
# #     assert len(seqs) == 221


# # def test_api_integration_bcr_hiv_bnab_lcs():
# #     test_data = os.path.abspath('abstar/test_data/test_hiv_bnab_lcs.fasta')
# #     with open(test_data, 'r') as f:
# #         test_seqs = [Sequence(s) for s in SeqIO.parse(f, 'fasta')]
# #     seqs = abstar.run(*test_seqs)
# #     assert len(seqs) == 207


# # @pytest.mark.parametrize('num_cores', [1, 4])
# # def test_chunks(num_cores):
# #     """Test that a small file is correctly split into chunks, and returns correct set of json files"""
# #     temp_dir = tempfile.mkdtemp()
# #     out_dir = tempfile.mkdtemp()
# #     arg_list = ['-i', 'abstar/test_data/test.fastq', '-o', out_dir, '-t', temp_dir, '-r', 'tcr', '--num-cores', str(num_cores), '--chunksize', '6']
# #     abstar.run_main(arg_list)
# #     assert len(os.listdir(os.path.join(temp_dir, 'input'))) == len(os.listdir(os.path.join(temp_dir, 'json')))



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


import logging
import os
import sys
from typing import Iterable, Optional, Union

import abutils
import polars as pl

from .assigner import AssignerBase


class MMseqs(AssignerBase):
    def __init__(
        self,
        output_directory: str,
        log_directory: str,
        germdb_name: str,
        receptor: str,
        logger: Optional[logging.Logger] = None,
        concise_logging: bool = False,
        chunksize: int = 1e6,
        threads: Optional[int] = None,
        debug: bool = False,
    ) -> None:
        """
        Initialize the MMseqs assigner.
        """
        super().__init__(
            output_directory=output_directory,
            log_directory=log_directory,
            germdb_name=germdb_name,
            receptor=receptor,
            logger=logger,
            concise_logging=concise_logging,
            debug=debug,
        )

        self.chunksize = chunksize
        self.threads = threads

    def __call__(self, sequence_file: str) -> str:
        """
        Run the MMseqs assigner.

        Parameters
        ----------
        sequence_file : str
            The path to the input file, in either FASTA or FASTQ format. Gzip-compressed files
            are supported.

        Returns
        -------
        str
            The path to the output Parquet file.
        """
        self.sample_name = ".".join(
            os.path.basename(sequence_file).rstrip(".gz").split(".")[:-1]
        )

        # input files are split into 1M sequence chunks
        input_fastas, input_tsvs, sequence_count = self.prepare_input_files(
            sequence_file, chunksize=self.chunksize
        )
        if self.concise_logging:
            self.logger.info(f"input sequences: {sequence_count:,}\n")
            self.logger.info("germline assignment: V | D | J | C")
        else:
            self.logger.info(f"input contained {sequence_count:,} sequences\n")
            self.logger.info("\n")
            self.logger.info("germline assignment:\n")
            self.logger.info("  V | D | J | C")

        if self.threads is None and sequence_count < 1000:
            self.threads = 1

        if self.debug:
            print(f"MMSEQS THREADS: {self.threads}")

        # process input data file(s)
        assigned_paths = []
        unassigned_paths = []
        for input_fasta, input_tsv in zip(input_fastas, input_tsvs):
            assigned, unassigned = self.assign_germlines(
                input_fasta=input_fasta, input_tsv=input_tsv
            )
            assigned_paths.append(assigned)
            unassigned_paths.append(unassigned)

        # concatenate assigned and unassigned files
        assigned_path = self.assemble_output_files(assigned_paths, unassigned_paths)

        return assigned_path, sequence_count

    def assign_germlines(self, input_fasta: str, input_tsv: str) -> str:
        """
        V, D, J, and C germline gene segment assignment.

        Parameters
        ----------
        input_fastas : Iterable[str]
            An iterable of path(s) to one or more input FASTA files.

        input_tsv : str
            The path to the input TSV file.

        Returns
        -------
        str
            The path to the output Parquet file.
        """
        mmseqs_format_output = "query,target,evalue,qstart,qend,qseq,nident"
        germdb_path = os.path.join(self.germdb_path, "mmseqs")
        input_df = pl.scan_csv(input_tsv, separator="\t")
        idx = input_fasta.split(".")[-2]

        # -----------
        #   V genes
        # -----------
        v_germdb = os.path.join(germdb_path, "v")

        # process input FASTA files
        vresult_path = os.path.join(
            self.output_directory, f"{self.sample_name}.vresult.{idx}.tsv"
        )
        if not self.debug:
            self.to_delete.append(vresult_path)
        # assign V genes
        abutils.tl.mmseqs_search(
            query=input_fasta,
            target=v_germdb,
            output_path=vresult_path,
            search_type=3,
            max_seqs=25,
            sensitivity=7.5,
            # max_evalue=1.0e-6,
            format_mode=4,
            additional_cli_args="--min-aln-len 12 -k 5 --alignment-mode 3",
            format_output=mmseqs_format_output,
            log_to=os.path.join(
                self.log_directory, f"{self.sample_name}.v_assignment.log"
            ),
            threads=self.threads,
            debug=self.debug,
        )
        # read the results
        vresult_df = pl.scan_csv(  # vresult_df is a LazyFrame
            vresult_path,
            separator="\t",
            with_column_names=lambda x: [
                f"v_{_x}".replace("target", "call").replace("evalue", "support")
                for _x in x
            ],
        )
        # keep only the highest scoring assignment for each sequence
        vresult_df = vresult_df.sort(
            by=["v_nident"],
            descending=True,
            nulls_last=True,
        )
        vresult_df = vresult_df.unique(subset=["v_query"], keep="first").collect()

        # -----------
        #   J genes
        # -----------

        # self.logger.info(" | J")
        j_germdb = os.path.join(germdb_path, "j")
        jquery_path = os.path.join(
            self.output_directory, f"{self.sample_name}.jquery.{idx}.fasta"
        )
        jresult_path = os.path.join(
            self.output_directory, f"{self.sample_name}.jresult.{idx}.tsv"
        )
        if not self.debug:
            self.to_delete.extend([jresult_path, jquery_path])
        # make the input FASTA file for J gene assignment
        # self.build_jquery_fasta(vresult_df, jquery_path)
        self.build_jquery_fasta(vresult_df, jquery_path)
        # assign J genes
        abutils.tl.mmseqs_search(
            query=jquery_path,
            target=j_germdb,
            output_path=jresult_path,
            search_type=3,
            max_seqs=25,
            max_evalue=1000.0,
            format_mode=4,
            additional_cli_args="--min-aln-len 12 -k 5 --alignment-mode 3",
            format_output=mmseqs_format_output,
            log_to=os.path.join(
                self.log_directory, f"{self.sample_name}.j_assignment.log"
            ),
            threads=self.threads,
            debug=self.debug,
        )
        # read the results
        jresult_df = pl.scan_csv(
            jresult_path,
            separator="\t",
            with_column_names=lambda x: [
                f"j_{_x}".replace("target", "call").replace("evalue", "support")
                for _x in x
            ],
        )

        # keep only the highest scoring assignment for each sequence
        jresult_df = jresult_df.sort(by=["j_nident"], descending=True, nulls_last=True)
        jresult_df = jresult_df.unique(subset=["j_query"], keep="first").collect()
        # join the V and J assignment results
        vjresult_df = vresult_df.join(
            jresult_df,
            left_on="v_query",
            right_on="j_query",
            how="left",
        )

        if self.debug:
            vjresult_df.write_csv(
                os.path.join(
                    self.output_directory, f"{self.sample_name}.vjresult.{idx}.csv"
                ),
            )

        # -----------
        #   D genes
        # -----------

        # self.logger.info(" | D")
        d_germdb = os.path.join(germdb_path, "d")
        dquery_path = os.path.join(
            self.output_directory, f"{self.sample_name}.dquery.{idx}.fasta"
        )
        dresult_path = os.path.join(
            self.output_directory, f"{self.sample_name}.dresult.{idx}.tsv"
        )
        if not self.debug:
            self.to_delete.extend([dresult_path, dquery_path])
        # make the input FASTA file for D gene assignment
        self.build_dquery_fasta(vjresult_df, dquery_path, minimum_length=5)

        # only assign D genes if the dquery fasta file is not empty
        # if os.path.exists(dquery_path):
        if abutils.io.determine_fastx_format(dquery_path) == "fasta":
            # assign D genes
            abutils.tl.mmseqs_search(
                query=dquery_path,
                target=d_germdb,
                output_path=dresult_path,
                search_type=3,
                max_seqs=25,
                max_evalue=1000000.0,
                format_mode=4,
                additional_cli_args="--min-aln-len 5 -k 3 --alignment-mode 3",
                format_output=mmseqs_format_output,
                log_to=os.path.join(
                    self.log_directory, f"{self.sample_name}.d_assignment.log"
                ),
                threads=self.threads,
                debug=self.debug,
            )
            # read the results
            dresult_df = pl.scan_csv(
                dresult_path,
                separator="\t",
                with_column_names=lambda x: [
                    f"d_{_x}".replace("target", "call").replace("evalue", "support")
                    for _x in x
                ],
            )
            # keep only the highest scoring assignment for each sequence
            dresult_df = dresult_df.sort(
                by=["d_nident"], descending=True, nulls_last=True
            )
            dresult_df = dresult_df.unique(subset=["d_query"], keep="first").collect()
            if dresult_df.shape[0] > 0:
                # join the D and VJ assignment results
                vdjresult_df = vjresult_df.join(
                    dresult_df,
                    left_on="v_query",
                    right_on="d_query",
                    how="left",
                )
            else:
                # if none of the sequences have a D gene assignment,
                # set the D gene columns to None
                vdjresult_df = vjresult_df.with_columns(
                    pl.lit(None).alias("d_call"),
                    pl.lit(None).alias("d_support"),
                )
        else:
            # if the dquery fasta file is empty,
            # set the D gene columns to None
            vdjresult_df = vjresult_df.with_columns(
                pl.lit(None).alias("d_call"),
                pl.lit(None).alias("d_support"),
            )

        if self.debug:
            vdjresult_df.write_csv(
                os.path.join(
                    self.output_directory, f"{self.sample_name}.vdjresult.{idx}.csv"
                ),
            )

        # -----------
        #   C genes
        # -----------
        c_germdb = os.path.join(germdb_path, "c")

        # some germline databases may not have constant genes
        # so we only try to assign if the database exists
        if os.path.exists(c_germdb):
            # self.logger.info(" | C")
            cquery_path = os.path.join(
                self.output_directory, f"{self.sample_name}.cquery.{idx}.fasta"
            )
            cresult_path = os.path.join(
                self.output_directory, f"{self.sample_name}.cresult.{idx}.tsv"
            )
            if not self.debug:
                self.to_delete.extend([cresult_path, cquery_path])
            # make the input FASTA file for C gene assignment
            self.build_cquery_fasta(vjresult_df, cquery_path)
            # assign C genes
            abutils.tl.mmseqs_search(
                query=cquery_path,
                target=c_germdb,
                output_path=cresult_path,
                search_type=3,
                # max_seqs=25,
                max_evalue=10.0,
                format_mode=4,
                additional_cli_args="--min-aln-len 12 -k 5 --alignment-mode 3",
                format_output=mmseqs_format_output,
                log_to=os.path.join(
                    self.log_directory, f"{self.sample_name}.c_assignment.log"
                ),
                threads=self.threads,
                debug=self.debug,
            )
            # read the results
            cresult_df = pl.scan_csv(
                cresult_path,
                separator="\t",
                with_column_names=lambda x: [
                    f"c_{_x}".replace("target", "call").replace("evalue", "support")
                    for _x in x
                ],
            )
            # keep only the highest scoring assignment for each sequence
            cresult_df = cresult_df.sort(
                by=["c_nident"], descending=True, nulls_last=True
            )
            cresult_df = cresult_df.unique(subset=["c_query"], keep="first").collect()
            if cresult_df.shape[0] > 0:
                # join the C and VDJ assignment results
                vdjcresult_df = vdjresult_df.join(
                    cresult_df,
                    left_on="v_query",
                    right_on="c_query",
                    how="left",
                )
            else:
                # if none of the sequences have a C gene assignment,
                # set the C gene columns to None
                vdjcresult_df = vdjresult_df.with_columns(
                    pl.lit(None).alias("c_call"),
                    pl.lit(None).alias("c_support"),
                )
        else:
            # if the C segment germline database does not exist,
            # set the C gene columns to None
            vdjcresult_df = vdjresult_df.with_columns(
                pl.lit(None).alias("c_call"),
                pl.lit(None).alias("c_support"),
            )

        input_df = pl.read_csv(input_tsv, separator="\t")
        vdjcresult_df = input_df.join(
            vdjcresult_df,
            left_on="sequence_id",
            right_on="v_query",
            how="left",
        )

        # add the rev_comp column
        vdjcresult_df = vdjcresult_df.with_columns(
            (pl.col("v_qstart") > pl.col("v_qend")).alias("rev_comp")
        )

        if self.debug:
            vdjcresult_df.write_csv(
                os.path.join(
                    self.output_directory, f"{self.sample_name}.vdjcresult.{idx}.csv"
                ),
            )

        # log "unassigned" sequences (no V or J gene assignment)
        unassigned_cols = ["sequence_id", "sequence_input"]
        unassigned = vdjcresult_df.filter(
            pl.col("v_call").is_null() | pl.col("j_call").is_null()
        )
        unassigned = unassigned.select(unassigned_cols)
        unassigned_path = os.path.join(
            self.log_directory, f"{self.sample_name}.{idx}.unassigned.csv"
        )
        unassigned.write_csv(unassigned_path)
        if not self.debug:
            self.to_delete.append(unassigned_path)

        # write "assigned" sequence results (successful V and J gene assignment)
        assigned_cols = [
            "sequence_id",
            "sequence_input",
            "quality",
            "rev_comp",
            "v_call",
            "v_support",
            "d_call",
            "d_support",
            "j_call",
            "j_support",
            "c_call",
            "c_support",
        ]
        assigned = vdjcresult_df.filter(
            ~pl.col("v_call").is_null() & ~pl.col("j_call").is_null()
        )
        assigned = assigned.select(assigned_cols)
        assigned_path = os.path.join(
            self.output_directory, f"{self.sample_name}.{idx}.parquet"
        )
        assigned.write_parquet(assigned_path)
        if not self.debug:
            self.to_delete.append(assigned_path)

        return assigned_path, unassigned_path

    def prepare_input_files(
        self, sequence_file: str, chunksize: int = 1e6
    ) -> Iterable[str]:
        """
        Prepare the input file for use by the Assigner.

        MMSeqs requires FASTA-formatted files, but the input file may be FASTQ-formatted and
        may also be gzip-compressed. This method handles the conversion and decompression of the
        input file.

        Parameters
        ----------
        sequence_file : str
            The path to the input file.

        Returns
        -------
        Iterable[Iterable[str], Iterable[str], int]
            An iterable containing the path to one or more FASTA-formatted input files, an iterable containing
            the path to one or more 3-column tab-delimited files containing the sequence ID, input sequence,
            and (optionally) quality score for each input sequence, and the total sequence count.

            .. note:
                The output format needs to be tab-separated rather than comma-separated because the comma
                is a valid character in quality scores and will obviously mess with CSV formatting.
        """
        # count input sequences
        sequence_count = 0
        for seq in abutils.io.parse_fastx(sequence_file):
            sequence_count += 1

        # split input file, if necessary
        if sequence_count > chunksize:
            sequence_files = abutils.io.split_fastx(
                sequence_file,
                output_directory=os.path.dirname(sequence_file),
                chunksize=chunksize,
            )
        else:
            sequence_files = [sequence_file]

        # process sequence file(s)
        output_fastas = []
        output_csvs = []
        for i, sequence_file in enumerate(sequence_files):
            # set up output files
            output_fasta = os.path.join(
                self.output_directory, f"{self.sample_name}.{i}.fasta"
            )
            output_csv = os.path.join(
                self.output_directory, f"{self.sample_name}.{i}.tsv"
            )
            if not self.debug:
                self.to_delete.extend([output_fasta, output_csv])

            # process input file
            with open(output_fasta, "w") as ofasta:
                with open(output_csv, "w") as ocsv:
                    ocsv.write("sequence_id\tsequence_input\tquality\n")  # header
                    for seq in abutils.io.parse_fastx(sequence_file):
                        qual = seq.qual if seq.qual is not None else ""
                        if qual:
                            if qual[0] in ['"', "'"]:  # leading quotes cause issues
                                qual = "#" + qual[1:]
                        ofasta.write(f">{seq.id}\n{seq.sequence}\n")
                        ocsv.write(f"{seq.id}\t{seq.sequence}\t{qual}\n")
            output_fastas.append(output_fasta)
            output_csvs.append(output_csv)

        # delete chunked sequence files (if there's only one file, it hasn't been chunked)
        if len(sequence_files) > 1:
            abutils.io.delete_files(sequence_files)

        return output_fastas, output_csvs, sequence_count

    def assemble_output_files(
        self,
        assigned_paths: Iterable[str],
        unassigned_paths: Iterable[str],
    ) -> Iterable[str]:
        """
        Assemble the output files. Uses Polars' ``scan`` and ``sink`` methods
        for memory efficiency.

        Parameters
        ----------
        assigned_paths : Iterable[str]
            The paths to the assigned Parquet files.

        unassigned_paths : Iterable[str]
            The paths to the unassigned CSV files.

        Returns
        -------
        str
            The path to the assembled Parquet file.

        """
        # assigned files
        assigned = os.path.join(self.output_directory, f"{self.sample_name}.parquet")
        assigned_dfs = []
        for assigned_path in assigned_paths:
            _df = pl.scan_parquet(assigned_path)
            assigned_dfs.append(_df)
        assigned_df = pl.concat(assigned_dfs)
        assigned_df.sink_parquet(assigned, statistics=False)

        # unassigned_files
        unassigned = os.path.join(
            self.log_directory, f"{self.sample_name}.unassigned.csv"
        )
        unassigned_dfs = []
        for unassigned_path in unassigned_paths:
            _df = pl.scan_csv(unassigned_path)
            unassigned_dfs.append(_df)
        unassigned_df = pl.concat(unassigned_dfs)
        unassigned_df.sink_csv(unassigned)

        return assigned

    def build_jquery_fasta(
        self,
        vresult_df: Union[pl.LazyFrame, pl.DataFrame],
        fasta_path: str,
    ) -> None:
        """
        Builds a FASTA file for the J gene assignment.
        The query sequence is the portion of the input sequence
        that follows the region assigned to the V gene.

        Parameters
        ----------
        vresult_df : pl.LazyFrame or pl.DataFrame
            The V gene assignment results.

        fasta_path : str
            The path to the output FASTA file.

        Returns
        -------
        None

        """
        vcols = ["v_query", "v_qstart", "v_qend", "v_qseq"]
        fastas = []
        if isinstance(vresult_df, pl.LazyFrame):
            vresult_df = vresult_df.collect()

        # retrieve a subset of the sequence that follows the V alignment
        for name, start, end, seq in vresult_df.select(
            pl.concat_list(pl.col(vcols)).alias("jq")
        )["jq"]:
            start = int(start)
            end = int(end)
            if start > end:
                s = seq[:end]
            else:
                s = seq[end:]
            # sequences shorter than 14 nt can break MMseqs2
            # when using our J-gene alignment paramters
            if len(s) >= 14:
                fastas.append(f">{name}\n{s}")

        # write the FASTA file
        with open(fasta_path, "w") as f:
            f.write("\n".join(fastas))

    def build_dquery_fasta(
        self,
        vjresult_df: Union[pl.LazyFrame, pl.DataFrame],
        fasta_path: str,
        minimum_length: int = 5,
    ) -> None:
        """
        Builds a FASTA file for the D gene assignment.
        The query sequence is the portion of the input sequence
        that falls between the regions assigned to V and J genes.

        Parameters
        ----------
        vjresult_df : pl.LazyFrame or pl.DataFrame
            The J gene assignment results.

        fasta_path : str
            The path to the output FASTA file.

        Returns
        -------
        None

        """
        vcols = ["v_query", "j_qstart", "j_qend", "j_qseq"]
        fastas = []

        # only assign D genes on "heavy" chains
        prefixes = ["IGH", "TRA", "TRD"]
        heavy_df = vjresult_df.filter(pl.col("v_call").str.contains_any(prefixes))
        heavy_df = heavy_df.filter(~pl.col("j_qstart").is_null())
        if isinstance(heavy_df, pl.LazyFrame):
            heavy_df = heavy_df.collect()

        # retrieve a subset of the sequence that preceeds the J alignment
        for name, start, end, seq in heavy_df.select(
            pl.concat_list(pl.col(vcols)).alias("dq")
        )["dq"]:
            start = int(start)
            end = int(end)
            if abs(start - end) < minimum_length:
                continue
            if start > end:
                fasta = f">{name}\n{seq[start:]}"
            else:
                fasta = f">{name}\n{seq[:start]}"
            fastas.append(fasta)

        # write the FASTA file
        with open(fasta_path, "w") as f:
            f.write("\n".join(fastas))

    def build_cquery_fasta(
        self,
        vjresult_df: Union[pl.LazyFrame, pl.DataFrame],
        fasta_path: str,
    ) -> None:
        """
        Builds a FASTA file for the C gene assignment.
        The query sequence is the portion of the input sequence
        that follows the region assigned to the J gene.

        Parameters
        ----------
        vjresult_df : pl.LazyFrame or pl.DataFrame
            The J gene assignment results.

        fasta_path : str
            The path to the output FASTA file.

        Returns
        -------
        None

        """
        vcols = ["v_query", "j_qstart", "j_qend", "j_qseq"]
        fastas = []
        if isinstance(vjresult_df, pl.LazyFrame):
            vjresult_df = vjresult_df.collect()
        vjresult_df = vjresult_df.drop_nulls(vcols)

        # retrieve a subset of the sequence that preceeds the J alignment
        for name, start, end, seq in vjresult_df.select(
            pl.concat_list(pl.col(vcols)).alias("cq")
        )["cq"]:
            start = int(start)
            end = int(end)
            if start > end:
                query = seq[:end]
            else:
                query = seq[end:]
            fasta = f">{name}\n{query}"
            fastas.append(fasta)

        # write the FASTA file
        with open(fasta_path, "w") as f:
            f.write("\n".join(fastas))


# class MMseqs(AssignerBase):
#     def __init__(
#         self,
#         output_directory: str,
#         log_directory: str,
#         germdb_name: str,
#         receptor: str,
#         logger: Optional[logging.Logger] = None,
#         concise_logging: bool = False,
#         debug: bool = False,
#     ) -> None:
#         """
#         Initialize the MMseqs assigner.
#         """
#         super().__init__(
#             output_directory=output_directory,
#             log_directory=log_directory,
#             germdb_name=germdb_name,
#             receptor=receptor,
#             logger=logger,
#             concise_logging=concise_logging,
#             debug=debug,
#         )

#     def __call__(self, sequence_file: str) -> str:
#         """
#         Run the MMseqs assigner.

#         Parameters
#         ----------
#         sequence_file : str
#             The path to the input file, in either FASTA or FASTQ format. Gzip-compressed files
#             are supported.

#         Returns
#         -------
#         str
#             The path to the output Parquet file.
#         """
#         self.sample_name = ".".join(
#             os.path.basename(sequence_file).rstrip(".gz").split(".")[:-1]
#         )
#         input_fasta, input_tsv, sequence_count = self.prepare_input_files(sequence_file)
#         if self.concise_logging:
#             self.logger.info(f"input sequences: {sequence_count:,}\n")
#         else:
#             self.logger.info(f"input contained {sequence_count:,} sequences\n")
#         assigned_path = self.assign_germlines(
#             input_fasta=input_fasta, input_tsv=input_tsv
#         )
#         return assigned_path, sequence_count

#     def assign_germlines(self, input_fasta: str, input_tsv: str) -> str:
#         """
#         V, D, J, and C germline gene segment assignment.

#         Parameters
#         ----------
#         input_fasta : str
#             The path to the input FASTA file.

#         input_tsv : str
#             The path to the input TSV file.

#         Returns
#         -------
#         str
#             The path to the output Parquet file.
#         """
#         mmseqs_format_output = "query,target,evalue,qstart,qend,qseq,nident"
#         germdb_path = os.path.join(self.germdb_path, "mmseqs")
#         input_df = pl.scan_csv(input_tsv, separator="\t")

#         # -----------
#         #   V genes
#         # -----------

#         if self.concise_logging:
#             self.logger.info("germline assignment: V")
#         else:
#             self.logger.info("\n")
#             self.logger.info("germline assignment:\n")
#             self.logger.info("  V")
#         v_germdb = os.path.join(germdb_path, "v")
#         vresult_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.vresult.tsv"
#         )
#         if not self.debug:
#             self.to_delete.append(vresult_path)
#         # assign V genes
#         abutils.tl.mmseqs_search(
#             query=input_fasta,
#             target=v_germdb,
#             output_path=vresult_path,
#             search_type=3,
#             max_seqs=25,
#             sensitivity=7.5,
#             # max_evalue=1.0e-6,
#             format_mode=4,
#             additional_cli_args="--min-aln-len 12 -k 5 --alignment-mode 3",
#             format_output=mmseqs_format_output,
#             log_to=os.path.join(self.log_directory, "v_assignment.log"),
#             debug=self.debug,
#         )
#         # read the results
#         vresult_df = pl.scan_csv(  # vresult_df is a LazyFrame
#             vresult_path,
#             separator="\t",
#             with_column_names=lambda x: [
#                 f"v_{_x}".replace("target", "call").replace("evalue", "support")
#                 for _x in x
#             ],
#         )
#         # keep only the highest scoring assignment for each sequence
#         # vresult_df = vresult_df.sort(by=["v_support"], nulls_last=True)
#         vresult_df = vresult_df.sort(by=["v_nident"], descending=True, nulls_last=True)
#         vresult_df = vresult_df.unique(subset=["v_query"], keep="first")

#         # -----------
#         #   J genes
#         # -----------

#         self.logger.info(" | J")
#         j_germdb = os.path.join(germdb_path, "j")
#         jquery_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.jquery.fasta"
#         )
#         jresult_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.jresult.tsv"
#         )
#         if not self.debug:
#             self.to_delete.extend([jresult_path, jquery_path])
#         # make the input FASTA file for J gene assignment
#         self.build_jquery_fasta(vresult_df, jquery_path)
#         # assign J genes
#         abutils.tl.mmseqs_search(
#             query=jquery_path,
#             target=j_germdb,
#             output_path=jresult_path,
#             search_type=3,
#             max_seqs=25,
#             max_evalue=1000.0,
#             format_mode=4,
#             additional_cli_args="--min-aln-len 12 -k 5 --alignment-mode 3",
#             format_output=mmseqs_format_output,
#             log_to=os.path.join(self.log_directory, "j_assignment.log"),
#             debug=self.debug,
#         )
#         # read the results
#         jresult_df = pl.scan_csv(
#             jresult_path,
#             separator="\t",
#             with_column_names=lambda x: [
#                 f"j_{_x}".replace("target", "call").replace("evalue", "support")
#                 for _x in x
#             ],
#         )
#         # keep only the highest scoring assignment for each sequence
#         jresult_df = jresult_df.sort(by=["j_nident"], descending=True, nulls_last=True)
#         jresult_df = jresult_df.unique(subset=["j_query"], keep="first")
#         # join the V and J assignment results
#         vjresult_df = vresult_df.join(
#             jresult_df,
#             left_on="v_query",
#             right_on="j_query",
#             how="left",
#         )

#         # -----------
#         #   D genes
#         # -----------

#         self.logger.info(" | D")
#         d_germdb = os.path.join(germdb_path, "d")
#         dquery_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.dquery.fasta"
#         )
#         dresult_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.dresult.tsv"
#         )
#         if not self.debug:
#             self.to_delete.extend([dresult_path, dquery_path])
#         # make the input FASTA file for D gene assignment
#         self.build_dquery_fasta(vjresult_df, dquery_path, minimum_length=5)

#         # only assign D genes if the dquery fasta file is not empty
#         # if os.path.exists(dquery_path):
#         if abutils.io.determine_fastx_format(dquery_path) == "fasta":
#             # assign D genes
#             abutils.tl.mmseqs_search(
#                 query=dquery_path,
#                 target=d_germdb,
#                 output_path=dresult_path,
#                 search_type=3,
#                 max_seqs=25,
#                 max_evalue=1000000.0,
#                 format_mode=4,
#                 additional_cli_args="--min-aln-len 5 -k 3 --alignment-mode 3",
#                 format_output=mmseqs_format_output,
#                 log_to=os.path.join(self.log_directory, "d_assignment.log"),
#                 debug=self.debug,
#             )
#             # read the results
#             dresult_df = pl.scan_csv(
#                 dresult_path,
#                 separator="\t",
#                 with_column_names=lambda x: [
#                     f"d_{_x}".replace("target", "call").replace("evalue", "support")
#                     for _x in x
#                 ],
#             )
#             # keep only the highest scoring assignment for each sequence
#             dresult_df = dresult_df.sort(
#                 by=["d_nident"], descending=True, nulls_last=True
#             )
#             dresult_df = dresult_df.unique(subset=["d_query"], keep="first")
#             if dresult_df.select(pl.len()).collect().item() > 0:
#                 # join the D and VJ assignment results
#                 vdjresult_df = vjresult_df.join(
#                     dresult_df,
#                     left_on="v_query",
#                     right_on="d_query",
#                     how="left",
#                 )
#             else:
#                 # if none of the sequences have a D gene assignment,
#                 # set the D gene columns to None
#                 vdjresult_df = vjresult_df.with_columns(
#                     pl.lit(None).alias("d_call"),
#                     pl.lit(None).alias("d_support"),
#                 )
#         else:
#             # if the dquery fasta file is empty,
#             # set the D gene columns to None
#             vdjresult_df = vjresult_df.with_columns(
#                 pl.lit(None).alias("d_call"),
#                 pl.lit(None).alias("d_support"),
#             )

#         # -----------
#         #   C genes
#         # -----------
#         c_germdb = os.path.join(germdb_path, "c")
#         cquery_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.cquery.fasta"
#         )
#         # some germline databases may not have constant genes
#         # so we only try to assign if the database exists
#         if os.path.exists(c_germdb):
#             self.logger.info(" | C")
#             cresult_path = os.path.join(
#                 self.output_directory, f"{self.sample_name}.cresult.tsv"
#             )
#             if not self.debug:
#                 self.to_delete.extend([cresult_path, cquery_path])
#             # make the input FASTA file for C gene assignment
#             self.build_cquery_fasta(vjresult_df, cquery_path)
#             # assign C genes
#             abutils.tl.mmseqs_search(
#                 query=cquery_path,
#                 target=c_germdb,
#                 output_path=cresult_path,
#                 search_type=3,
#                 # max_seqs=25,
#                 max_evalue=10.0,
#                 format_mode=4,
#                 additional_cli_args="--min-aln-len 12 -k 5 --alignment-mode 3",
#                 format_output=mmseqs_format_output,
#                 log_to=os.path.join(self.log_directory, "c_assignment.log"),
#                 debug=self.debug,
#             )
#             # read the results
#             cresult_df = pl.scan_csv(
#                 cresult_path,
#                 separator="\t",
#                 with_column_names=lambda x: [
#                     f"c_{_x}".replace("target", "call").replace("evalue", "support")
#                     for _x in x
#                 ],
#             )
#             # keep only the highest scoring assignment for each sequence
#             cresult_df = cresult_df.sort(
#                 by=["c_nident"], descending=True, nulls_last=True
#             )
#             cresult_df = cresult_df.unique(subset=["c_query"], keep="first")
#             if cresult_df.select(pl.len()).collect().item() > 0:
#                 # join the C and VDJ assignment results
#                 vdjcresult_df = vdjresult_df.join(
#                     cresult_df,
#                     left_on="v_query",
#                     right_on="c_query",
#                     how="left",
#                 )
#             else:
#                 # if none of the sequences have a C gene assignment,
#                 # set the C gene columns to None
#                 vdjcresult_df = vdjresult_df.with_columns(
#                     pl.lit(None).alias("c_call"),
#                     pl.lit(None).alias("c_support"),
#                 )
#         else:
#             # if the C segment germline database does not exist,
#             # set the C gene columns to None
#             vdjcresult_df = vdjresult_df.with_columns(
#                 pl.lit(None).alias("c_call"),
#                 pl.lit(None).alias("c_support"),
#             )

#         # join the input CSV data with VDJC assignment results
#         if self.concise_logging:
#             self.logger.info("\n")
#         else:
#             self.logger.info("\n")
#             self.logger.info("combining germline assignment results\n")
#         vdjcresult_df = input_df.join(
#             vdjcresult_df,
#             left_on="sequence_id",
#             right_on="v_query",
#             how="left",
#         )

#         # add the rev_comp column
#         vdjcresult_df = vdjcresult_df.with_columns(
#             (pl.col("v_qstart") > pl.col("v_qend")).alias("rev_comp")
#         )

#         # log "unassigned" sequences (no V or J gene assignment)
#         unassigned_cols = ["sequence_id", "sequence_input"]
#         unassigned = vdjcresult_df.filter(
#             pl.col("v_call").is_null() | pl.col("j_call").is_null()
#         )
#         unassigned = unassigned.select(unassigned_cols)
#         unassigned_path = os.path.join(
#             self.log_directory, f"{self.sample_name}.unassigned.csv"
#         )
#         unassigned.sink_csv(unassigned_path)

#         # write "assigned" sequence results (successful V and J gene assignment)
#         assigned_cols = [
#             "sequence_id",
#             "sequence_input",
#             "quality",
#             "rev_comp",
#             "v_call",
#             "v_support",
#             "d_call",
#             "d_support",
#             "j_call",
#             "j_support",
#             "c_call",
#             "c_support",
#         ]
#         assigned = vdjcresult_df.filter(
#             ~pl.col("v_call").is_null() & ~pl.col("j_call").is_null()
#         )
#         assigned = assigned.select(assigned_cols)
#         assigned_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.parquet"
#         )
#         assigned.sink_parquet(assigned_path)

#         return assigned_path

#     def prepare_input_files(self, sequence_file: str) -> Iterable[str]:
#         """
#         Prepare the input file for use by the Assigner.

#         MMSeqs requires FASTA-formatted files, but the input file may be FASTQ-formatted and
#         may also be gzip-compressed. This method handles the conversion and decompression of the
#         input file.

#         Parameters
#         ----------
#         sequence_file : str
#             The path to the input file.

#         Returns
#         -------
#         Iterable[str]
#             An iterable containing the path to a FASTA-formatted input file, and the path to a 3-column
#             tab-delimited file containing the sequence ID, input sequence, and (optionally) quality score
#             for each input sequence.

#             .. note:
#                 The output format needs to be tab-separated rather than comma-separated because the comma
#                 is a valid character in quality scores and will obviously mess with CSV formatting.
#         """
#         # set up output files
#         output_fasta = os.path.join(self.output_directory, f"{self.sample_name}.fasta")
#         output_csv = os.path.join(self.output_directory, f"{self.sample_name}.tsv")
#         if not self.debug:
#             self.to_delete.extend([output_fasta, output_csv])

#         # process input file
#         sequence_count = 0
#         with open(output_fasta, "w") as ofasta:
#             with open(output_csv, "w") as ocsv:
#                 ocsv.write("sequence_id\tsequence_input\tquality\n")  # header
#                 for seq in abutils.io.parse_fastx(sequence_file):
#                     qual = seq.qual if seq.qual is not None else ""
#                     ofasta.write(f">{seq.id}\n{seq.sequence}\n")
#                     ocsv.write(f"{seq.id}\t{seq.sequence}\t{qual}\n")
#                     sequence_count += 1
#         return output_fasta, output_csv, sequence_count

#     def build_jquery_fasta(
#         self,
#         vresult_df: Union[pl.LazyFrame, pl.DataFrame],
#         fasta_path: str,
#     ) -> None:
#         """
#         Builds a FASTA file for the J gene assignment.
#         The query sequence is the portion of the input sequence
#         that follows the region assigned to the V gene.

#         Parameters
#         ----------
#         vresult_df : pl.LazyFrame or pl.DataFrame
#             The V gene assignment results.

#         fasta_path : str
#             The path to the output FASTA file.

#         Returns
#         -------
#         None

#         """
#         vcols = ["v_query", "v_qstart", "v_qend", "v_qseq"]
#         fastas = []
#         if isinstance(vresult_df, pl.LazyFrame):
#             vresult_df = vresult_df.collect()

#         # retrieve a subset of the sequence that follows the V alignment
#         for name, start, end, seq in vresult_df.select(
#             pl.concat_list(pl.col(vcols)).alias("jq")
#         )["jq"]:
#             start = int(start)
#             end = int(end)
#             if start > end:
#                 s = seq[:end]
#             else:
#                 s = seq[end:]
#             # sequences shorter than 14 nt can break MMseqs2
#             # when using our J-gene alignment paramters
#             if len(s) >= 14:
#                 fastas.append(f">{name}\n{s}")

#         # write the FASTA file
#         with open(fasta_path, "w") as f:
#             f.write("\n".join(fastas))

#     def build_dquery_fasta(
#         self,
#         vjresult_df: Union[pl.LazyFrame, pl.DataFrame],
#         fasta_path: str,
#         minimum_length: int = 5,
#     ) -> None:
#         """
#         Builds a FASTA file for the D gene assignment.
#         The query sequence is the portion of the input sequence
#         that falls between the regions assigned to V and J genes.

#         Parameters
#         ----------
#         vjresult_df : pl.LazyFrame or pl.DataFrame
#             The J gene assignment results.

#         fasta_path : str
#             The path to the output FASTA file.

#         Returns
#         -------
#         None

#         """
#         vcols = ["v_query", "j_qstart", "j_qend", "j_qseq"]
#         fastas = []

#         # only assign D genes on "heavy" chains
#         prefixes = ["IGH", "TRA", "TRD"]
#         heavy_df = vjresult_df.filter(pl.col("v_call").str.contains_any(prefixes))
#         heavy_df = heavy_df.filter(~pl.col("j_qstart").is_null())
#         if isinstance(heavy_df, pl.LazyFrame):
#             heavy_df = heavy_df.collect()

#         # retrieve a subset of the sequence that preceeds the J alignment
#         for name, start, end, seq in heavy_df.select(
#             pl.concat_list(pl.col(vcols)).alias("dq")
#         )["dq"]:
#             start = int(start)
#             end = int(end)
#             if abs(start - end) < minimum_length:
#                 continue
#             if start > end:
#                 fasta = f">{name}\n{seq[start:]}"
#             else:
#                 fasta = f">{name}\n{seq[:start]}"
#             fastas.append(fasta)

#         # write the FASTA file
#         with open(fasta_path, "w") as f:
#             f.write("\n".join(fastas))

#     def build_cquery_fasta(
#         self,
#         vjresult_df: Union[pl.LazyFrame, pl.DataFrame],
#         fasta_path: str,
#     ) -> None:
#         """
#         Builds a FASTA file for the C gene assignment.
#         The query sequence is the portion of the input sequence
#         that follows the region assigned to the J gene.

#         Parameters
#         ----------
#         vjresult_df : pl.LazyFrame or pl.DataFrame
#             The J gene assignment results.

#         fasta_path : str
#             The path to the output FASTA file.

#         Returns
#         -------
#         None

#         """
#         vcols = ["v_query", "j_qstart", "j_qend", "j_qseq"]
#         fastas = []
#         if isinstance(vjresult_df, pl.LazyFrame):
#             vjresult_df = vjresult_df.collect()
#         vjresult_df = vjresult_df.drop_nulls(vcols)

#         # retrieve a subset of the sequence that preceeds the J alignment
#         for name, start, end, seq in vjresult_df.select(
#             pl.concat_list(pl.col(vcols)).alias("cq")
#         )["cq"]:
#             start = int(start)
#             end = int(end)
#             if start > end:
#                 query = seq[:end]
#             else:
#                 query = seq[end:]
#             fasta = f">{name}\n{query}"
#             fastas.append(fasta)

#         # write the FASTA file
#         with open(fasta_path, "w") as f:
#             f.write("\n".join(fastas))


#
#
#


# class MMseqs(AssignerBase):
#     def __init__(
#         self,
#         output_directory: str,
#         log_directory: str,
#         germdb_name: str,
#         receptor: str,
#         verbose: bool = False,
#         debug: bool = False,
#     ) -> None:
#         """
#         Initialize the MMseqs assigner.
#         """
#         super().__init__(
#             output_directory=output_directory,
#             log_directory=log_directory,
#             germdb_name=germdb_name,
#             receptor=receptor,
#             verbose=verbose,
#             debug=debug,
#         )

#     def __call__(self, sequence_file: str) -> str:
#         """
#         Run the MMseqs assigner.

#         Parameters
#         ----------
#         sequence_file : str
#             The path to the input file, in either FASTA or FASTQ format. Gzip-compressed files
#             are supported.

#         Returns
#         -------
#         str
#             The path to the output Parquet file.
#         """
#         self.sample_name = ".".join(
#             os.path.basename(sequence_file).rstrip(".gz").split(".")[:-1]
#         )

#         if self.verbose:
#             print("  - preparing input file")
#         input_fasta, input_tsv = self.prepare_input_files(sequence_file)
#         return self.assign_germlines(input_fasta=input_fasta, input_tsv=input_tsv)

#     def assign_germlines(self, input_fasta: str, input_tsv: str) -> str:
#         """
#         V, D and J germline gene segment assignment.

#         Parameters
#         ----------
#         input_fasta : str
#             The path to the input FASTA file.

#         input_tsv : str
#             The path to the input TSV file.

#         Returns
#         -------
#         str
#             The path to the output Parquet file.
#         """
#         mmseqs_format_output = "query,target,evalue,qstart,qend,qseq"
#         germdb_path = os.path.join(self.germdb_path, "mmseqs")
#         input_df = pl.scan_csv(input_tsv, separator="\t")

#         # -----------
#         #   V genes
#         # -----------
#         if self.verbose:
#             print("  - assigning V genes")
#         v_germdb = os.path.join(germdb_path, "v")
#         vresult_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.vresult.tsv"
#         )
#         if not self.debug:
#             self.to_delete.append(vresult_path)
#         # assign V genes
#         abutils.tl.mmseqs_search(
#             query=input_fasta,
#             target=v_germdb,
#             output_path=vresult_path,
#             search_type=3,
#             max_seqs=25,
#             max_evalue=1.0e-6,
#             format_mode=4,
#             format_output=mmseqs_format_output,
#             debug=self.debug,
#         )
#         # read the results
#         vresult_df = pl.scan_csv(  # vresult_df is a LazyFrame
#             vresult_path,
#             separator="\t",
#             # truncate_ragged_lines=True,
#             with_column_names=lambda x: [
#                 f"v_{_x}".replace("target", "call") for _x in x
#             ],
#         )
#         # keep only the highest scoring assignment for each sequence
#         vresult_df = vresult_df.sort(by=["v_evalue"], nulls_last=True)
#         vresult_df = vresult_df.unique(subset=["v_query"], keep="first")

#         # -----------
#         #   J genes
#         # -----------
#         if self.verbose:
#             print("  - assigning J genes")
#         j_germdb = os.path.join(germdb_path, "j")
#         jquery_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.jquery.fasta"
#         )
#         jresult_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.jresult.tsv"
#         )
#         if not self.debug:
#             self.to_delete.extend([jresult_path, jquery_path])
#         # make the input FASTA file for J gene assignment
#         self.build_jquery_fasta(vresult_df, jquery_path)
#         # assign J genes
#         abutils.tl.mmseqs_search(
#             query=jquery_path,
#             target=j_germdb,
#             output_path=jresult_path,
#             search_type=3,
#             max_seqs=25,
#             max_evalue=1000.0,
#             format_mode=4,
#             additional_cli_args="--min-aln-len 12 -k 3",
#             format_output=mmseqs_format_output,
#             debug=self.debug,
#         )
#         # read the results
#         jresult_df = pl.scan_csv(
#             jresult_path,
#             separator="\t",
#             # truncate_ragged_lines=True,
#             with_column_names=lambda x: [
#                 f"j_{_x}".replace("target", "call") for _x in x
#             ],
#         )
#         # keep only the highest scoring assignment for each sequence
#         jresult_df = jresult_df.sort(by=["j_evalue"], nulls_last=True)
#         jresult_df = jresult_df.unique(subset=["j_query"], keep="first")

#         # join the V and J assignment results
#         vjresult_df = vresult_df.join(
#             jresult_df,
#             left_on="v_query",
#             right_on="j_query",
#             how="left",
#         )

#         # -----------
#         #   D genes
#         # -----------
#         if self.verbose:
#             print("  - assigning D genes")
#         d_germdb = os.path.join(germdb_path, "d")
#         dquery_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.dquery.fasta"
#         )
#         dresult_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.dresult.tsv"
#         )
#         if not self.debug:
#             self.to_delete.extend([dresult_path, dquery_path])
#         # make the input FASTA file for D gene assignment
#         self.build_dquery_fasta(vjresult_df, dquery_path)
#         # assign D genes
#         abutils.tl.mmseqs_search(
#             query=dquery_path,
#             target=d_germdb,
#             output_path=dresult_path,
#             search_type=3,
#             max_seqs=25,
#             max_evalue=10000.0,
#             format_mode=4,
#             additional_cli_args="--min-aln-len 5 -k 3",
#             format_output=mmseqs_format_output,
#             debug=self.debug,
#         )
#         # read the results
#         dresult_df = pl.scan_csv(
#             dresult_path,
#             separator="\t",
#             # truncate_ragged_lines=True,
#             with_column_names=lambda x: [
#                 f"d_{_x}".replace("target", "call") for _x in x
#             ],
#         )
#         # keep only the highest scoring assignment for each sequence
#         dresult_df = dresult_df.sort(by=["d_evalue"], nulls_last=True)
#         dresult_df = dresult_df.unique(subset=["d_query"], keep="first")

#         # join the D and VJ assignment results
#         vdjresult_df = vjresult_df.join(
#             dresult_df,
#             left_on="v_query",
#             right_on="d_query",
#             how="left",
#         )

#         # join the input CSV data with VDJ assignment results
#         vdjresult_df = input_df.join(
#             vdjresult_df,
#             left_on="sequence_id",
#             right_on="v_query",
#             how="left",
#         )

#         # add the is_rc column
#         vdjresult_df = vdjresult_df.with_columns(
#             (pl.col("v_qstart") > pl.col("v_qend")).alias("is_rc")
#         )

#         # log "unassigned" sequences (no V or J gene assignment)
#         unassigned_cols = ["sequence_id", "sequence_input"]
#         unassigned = vdjresult_df.filter(
#             pl.col("v_call").is_null() | pl.col("j_call").is_null()
#         )
#         unassigned = unassigned.select(unassigned_cols)
#         unassigned_path = os.path.join(
#             self.log_directory, f"{self.sample_name}.unassigned.csv"
#         )
#         unassigned.sink_csv(unassigned_path)

#         # write "assigned" sequence results (successful V and J gene assignment)
#         assigned_cols = [
#             "sequence_id",
#             "sequence_input",
#             "quality",
#             "is_rc",
#             "v_call",
#             "v_evalue",
#             "d_call",
#             "d_evalue",
#             "j_call",
#             "j_evalue",
#         ]
#         assigned = vdjresult_df.filter(
#             ~pl.col("v_call").is_null() & ~pl.col("j_call").is_null()
#         )
#         assigned = assigned.select(assigned_cols)
#         assigned_path = os.path.join(
#             self.output_directory, f"{self.sample_name}.parquet"
#         )
#         assigned.sink_parquet(assigned_path)

#         return assigned_path

#     def prepare_input_files(self, sequence_file: str) -> Iterable[str]:
#         """
#         Prepare the input file for use by the Assigner.

#         MMSeqs requires FASTA-formatted files, but the input file may be FASTQ-formatted and
#         may also be gzip-compressed. This method handles the conversion and decompression of the
#         input file.

#         Parameters
#         ----------
#         sequence_file : str
#             The path to the input file.

#         Returns
#         -------
#         Iterable[str]
#             An iterable containing the path to a FASTA-formatted input file, and the path to a 3-column
#             tab-delimited file containing the sequence ID, input sequence, and (optionally) quality score
#             for each input sequence.

#             .. note:
#                 The output format needs to be tab-separated rather than comma-separated because the comma
#                 is a valid character in quality scores and will obviously mess with CSV formatting.
#         """
#         # set up output files
#         output_fasta = os.path.join(self.output_directory, f"{self.sample_name}.fasta")
#         output_csv = os.path.join(self.output_directory, f"{self.sample_name}.tsv")
#         if not self.debug:
#             self.to_delete.extend([output_fasta, output_csv])

#         # process input file
#         with open(output_fasta, "w") as ofasta:
#             with open(output_csv, "w") as ocsv:
#                 ocsv.write("sequence_id\tsequence_input\tquality\n")  # header
#                 for seq in abutils.io.parse_fastx(sequence_file):
#                     qual = seq.qual if seq.qual is not None else ""
#                     ofasta.write(f">{seq.id}\n{seq.sequence}\n")
#                     ocsv.write(f"{seq.id}\t{seq.sequence}\t{qual}\n")
#         return output_fasta, output_csv

#     def build_jquery_fasta(
#         self,
#         vresult_df: Union[pl.LazyFrame, pl.DataFrame],
#         fasta_path: str,
#     ) -> None:
#         """ """
#         vcols = ["v_query", "v_qstart", "v_qend", "v_qseq"]
#         fastas = []
#         if isinstance(vresult_df, pl.LazyFrame):
#             vresult_df = vresult_df.collect()

#         # retrieve a subset of the sequence that follows the V alignment
#         for name, start, end, seq in vresult_df.select(
#             pl.concat_list(pl.col(vcols)).alias("jq")
#         )["jq"]:
#             start = int(start)
#             end = int(end)
#             if start > end:
#                 fasta = f">{name}\n{seq[:end]}"
#             else:
#                 fasta = f">{name}\n{seq[end:]}"
#             fastas.append(fasta)

#         # write the FASTA file
#         with open(fasta_path, "w") as f:
#             f.write("\n".join(fastas))

#     def build_dquery_fasta(
#         self,
#         vjresult_df: Union[pl.LazyFrame, pl.DataFrame],
#         fasta_path: str,
#     ) -> None:
#         """ """
#         vcols = ["v_query", "j_qstart", "j_qend", "j_qseq"]
#         fastas = []

#         # only assign D genes on "heavy" chains
#         prefixes = ["IGH", "TRA", "TRD"]
#         heavy_df = vjresult_df.filter(pl.col("v_call").str.contains_any(prefixes))
#         heavy_df = heavy_df.filter(~pl.col("j_qstart").is_null())
#         if isinstance(heavy_df, pl.LazyFrame):
#             heavy_df = heavy_df.collect()

#         # retrieve a subset of the sequence that preceeds the J alignment
#         for name, start, end, seq in heavy_df.select(
#             pl.concat_list(pl.col(vcols)).alias("dq")
#         )["dq"]:
#             start = int(start)
#             end = int(end)
#             if start > end:
#                 fasta = f">{name}\n{seq[start:]}"
#             else:
#                 fasta = f">{name}\n{seq[:start]}"
#             fastas.append(fasta)

#         # write the FASTA file
#         with open(fasta_path, "w") as f:
#             f.write("\n".join(fastas))



# Copyright (c) 2024 Bryan Briney
# Distributed under the terms of the MIT License.
# SPDX-License-Identifier: MIT


from .antibody import Antibody
from .positions import (
    get_gapped_position_from_raw,
    get_raw_position_from_aligned,
    get_raw_position_from_gapped,
)


def get_region_sequence(
    region: str,
    aligned_sequence: str,
    aligned_germline: str,
    gapped_germline: str,
    germline_start: int,
    ab: Antibody,
    aa: bool = False,
) -> str:
    """
    Get the sequence of a region from the aligned sequence.

    Parameters
    ----------
    region : str
        The region to get the sequence of.

    aligned_sequence : str
        The aligned sequence to get the region from.

    aligned_germline : str
        The aligned germline sequence to get the region from.

    gapped_germline : str
        The gapped germline sequence to get the region from.

    germline_start : int
        The start position of the germline sequence.

    ab : Antibody
        The ``Antibody`` object. Used only for logging. No ``Antibody`` paremeters
        are required or updated.

    aa : bool, default False
        Whether to get the region in amino acids or nucleotides.

    Returns
    -------
    str
        The sequence of the region.

    """
    # IMGT start/end positions for the region
    if aa:
        imgt_start = IMGT_REGION_START_POSITIONS_AA[region]
        imgt_end = IMGT_REGION_END_POSITIONS_AA[region]
    else:
        imgt_start = IMGT_REGION_START_POSITIONS_NT[region]
        imgt_end = IMGT_REGION_END_POSITIONS_NT[region]
    ab.log(f"{region.upper()} IMGT START:", imgt_start)
    ab.log(f"{region.upper()} IMGT END:", imgt_end)

    # if the input sequence is sufficiently truncated at
    # the 5' end that the region is entirely missing...
    gapped_germline_start = get_gapped_position_from_raw(
        position=germline_start + 1,  # needs to be 1-indexed
        gapped_germline=gapped_germline,
    )
    if gapped_germline_start > imgt_end:
        return ""

    # positions in the aligned germline
    aligned_start = get_raw_position_from_gapped(
        position=imgt_start,
        gapped_germline=gapped_germline,
        germline_start=germline_start,
    )
    aligned_end = get_raw_position_from_gapped(
        position=imgt_end,
        gapped_germline=gapped_germline,
        germline_start=germline_start,
    )
    ab.log(f"{region.upper()} ALIGNED START:", aligned_start)
    ab.log(f"{region.upper()} ALIGNED END:", aligned_end)

    # positions in the unaligned sequence
    start = get_raw_position_from_aligned(
        position=aligned_start + 1,  # needs to be 1-indexed
        aligned_sequence=aligned_sequence,
        aligned_reference=aligned_germline,
    )
    end = get_raw_position_from_aligned(
        position=aligned_end + 1,  # needs to be 1-indexed
        aligned_sequence=aligned_sequence,
        aligned_reference=aligned_germline,
    )
    if start is not None:
        ab.log(f"{region.upper()} RAW START:", start - 1)  # actual slicing start
    else:
        ab.log(f"{region.upper()} RAW START: not found")
    if end is not None:
        ab.log(f"{region.upper()} RAW END:", end)
    else:
        ab.log(f"{region.upper()} RAW END: not found")

    # collect the sequence
    if start is not None and end is not None:
        region_sequence = aligned_sequence[start - 1 : end].replace("-", "")
    else:
        region_sequence = ""
    return region_sequence


# NOTE: these are the the actual IMGT end positions which are 1-indexed and
# not suitable for slicing in Python. To use these end points in a slice, subtract 1
# from the start position and use the end position unchanged, like so:
# ```
# start = IMGT_REGION_START_POSITIONS_NT["fwr1"] - 1
# end = IMGT_REGION_END_POSITIONS_NT["fwr1"]
# fwr1 = seq[start:end]
# ```
IMGT_REGION_END_POSITIONS_AA = {
    "fwr1": 26,
    "cdr1": 38,
    "fwr2": 55,
    "cdr2": 65,
    "fwr3": 104,
    "cdr3": 117,
    "fwr4": 129,
}

IMGT_REGION_END_POSITIONS_NT = {
    "fwr1": 78,
    "cdr1": 114,
    "fwr2": 165,
    "cdr2": 195,
    "fwr3": 312,
    "cdr3": 351,
    "fwr4": 387,
}

IMGT_REGION_START_POSITIONS_AA = {
    "fwr1": 1,
    "cdr1": 27,
    "fwr2": 39,
    "cdr2": 56,
    "fwr3": 66,
    "cdr3": 105,
    "fwr4": 118,
}

IMGT_REGION_START_POSITIONS_NT = {
    "fwr1": 1,
    "cdr1": 79,
    "fwr2": 115,
    "cdr2": 166,
    "fwr3": 196,
    "cdr3": 313,
    "fwr4": 352,
}
